"""synthetics base URL and path."""

from .responses import SyntheticsResponse

url_bases = [
    r"https?://synthetics\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/$": SyntheticsResponse.dispatch,
    "{0}/canary$": SyntheticsResponse.dispatch,
    "{0}/canary/(?P<name>[^/]+)$": SyntheticsResponse.dispatch,
    "{0}/canaries$": SyntheticsResponse.dispatch,
    "{0}/tags/(?P<resourceArn>[^/]+)$": SyntheticsResponse.dispatch,
}



import re
from re import Pattern
from typing import Any

from moto.core.parsers import XFormedDict
from moto.core.responses import ActionResult, BaseResponse, EmptyResult, PaginatedResult

from .exceptions import ValidationException
from .models import ElasticMapReduceBackend, emr_backends
from .utils import ReleaseLabel


class ElasticMapReduceResponse(BaseResponse):
    # EMR end points are inconsistent in the placement of region name
    # in the URL, so parsing it out needs to be handled differently
    emr_region_regex: list[Pattern[str]] = [
        re.compile(r"elasticmapreduce\.(.+?)\.amazonaws\.com"),
        re.compile(r"(.+?)\.elasticmapreduce\.amazonaws\.com"),
    ]
    PROTOCOL_PARSER_MAP_TYPE = XFormedDict
    RESPONSE_KEY_PATH_TO_TRANSFORMER = {
        "Properties": lambda x: x.original_dict() if hasattr(x, "original_dict") else x,
    }

    def __init__(self) -> None:
        super().__init__(service_name="emr")
        self.automated_parameter_parsing = True

    def get_region_from_url(self, request: Any, full_url: str) -> str:
        for regex in ElasticMapReduceResponse.emr_region_regex:
            match = regex.search(self.parsed_url.netloc)
            if match:
                return match.group(1)
        return self.default_region

    @property
    def backend(self) -> ElasticMapReduceBackend:
        return emr_backends[self.current_account][self.region]

    def add_instance_groups(self) -> ActionResult:
        jobflow_id = self._get_param("JobFlowId")
        instance_groups = self._get_param("InstanceGroups", [])
        fake_groups = self.backend.add_instance_groups(jobflow_id, instance_groups)
        result = {"InstanceGroups": fake_groups}
        return ActionResult(result)

    def add_job_flow_steps(self) -> ActionResult:
        job_flow_id = self._get_param("JobFlowId")
        steps = self.backend.add_job_flow_steps(
            job_flow_id, self._get_param("Steps", [])
        )
        result = {"StepIds": [step.id for step in steps]}
        return ActionResult(result)

    def add_tags(self) -> ActionResult:
        cluster_id = self._get_param("ResourceId")
        tags = self._get_param("Tags", [])
        tags = {d["Key"]: d["Value"] for d in tags}
        self.backend.add_tags(cluster_id, tags)
        return EmptyResult()

    def create_security_configuration(self) -> ActionResult:
        name = self._get_param("Name")
        security_configuration = self._get_param("SecurityConfiguration")
        security_configuration = self.backend.create_security_configuration(
            name=name, security_configuration=security_configuration
        )
        return ActionResult(security_configuration)

    def describe_security_configuration(self) -> ActionResult:
        name = self._get_param("Name")
        security_configuration = self.backend.get_security_configuration(name=name)
        return ActionResult(security_configuration)

    def delete_security_configuration(self) -> ActionResult:
        name = self._get_param("Name")
        self.backend.delete_security_configuration(name=name)
        return EmptyResult()

    def describe_cluster(self) -> ActionResult:
        cluster_id = self._get_param("ClusterId")
        cluster = self.backend.describe_cluster(cluster_id)
        result = {"Cluster": cluster}
        return ActionResult(result)

    def describe_job_flows(self) -> ActionResult:
        created_after = self._get_param("CreatedAfter")
        created_before = self._get_param("CreatedBefore")
        job_flow_ids = self._get_param("JobFlowIds", [])
        job_flow_states = self._get_param("JobFlowStates", [])
        clusters = self.backend.describe_job_flows(
            job_flow_ids, job_flow_states, created_after, created_before
        )
        result = {"JobFlows": clusters}
        return ActionResult(result)

    def describe_step(self) -> ActionResult:
        cluster_id = self._get_param("ClusterId")
        step_id = self._get_param("StepId")
        step = self.backend.describe_step(cluster_id, step_id)
        result = {"Step": step}
        return ActionResult(result)

    def list_bootstrap_actions(self) -> ActionResult:
        cluster_id = self._get_param("ClusterId")
        bootstrap_actions = self.backend.list_bootstrap_actions(cluster_id)
        result = {"BootstrapActions": bootstrap_actions}
        return PaginatedResult(result)

    def list_clusters(self) -> ActionResult:
        cluster_states = self._get_param("ClusterStates", [])
        created_after = self._get_param("CreatedAfter")
        created_before = self._get_param("CreatedBefore")
        clusters = self.backend.list_clusters(
            cluster_states, created_after, created_before
        )
        result = {"Clusters": clusters}
        return PaginatedResult(result)

    def list_instance_groups(self) -> ActionResult:
        cluster_id = self._get_param("ClusterId")
        instance_groups = self.backend.list_instance_groups(cluster_id)
        result = {"InstanceGroups": instance_groups}
        return PaginatedResult(result)

    def list_instances(self) -> ActionResult:
        cluster_id = self._get_param("ClusterId")
        instance_group_id = self._get_param("InstanceGroupId")
        instance_group_types = self._get_param("InstanceGroupTypes")
        instances = self.backend.list_instances(
            cluster_id,
            instance_group_id=instance_group_id,
            instance_group_types=instance_group_types,
        )
        result = {"Instances": instances}
        return PaginatedResult(result)

    def list_steps(self) -> ActionResult:
        cluster_id = self._get_param("ClusterId")
        step_ids = self._get_param("StepIds", [])
        step_states = self._get_param("StepStates", [])
        steps = self.backend.list_steps(
            cluster_id, step_ids=step_ids, step_states=step_states
        )
        result = {"Steps": steps}
        return PaginatedResult(result)

    def modify_cluster(self) -> ActionResult:
        cluster_id = self._get_param("ClusterId")
        step_concurrency_level = self._get_param("StepConcurrencyLevel")
        cluster = self.backend.modify_cluster(cluster_id, step_concurrency_level)
        result = {"StepConcurrencyLevel": cluster.step_concurrency_level}
        return ActionResult(result)

    def modify_instance_groups(self) -> ActionResult:
        instance_groups = self._get_param("InstanceGroups", [])
        self.backend.modify_instance_groups(instance_groups)
        return EmptyResult()

    def remove_tags(self) -> ActionResult:
        cluster_id = self._get_param("ResourceId")
        tag_keys = self._get_param("TagKeys", [])
        self.backend.remove_tags(cluster_id, tag_keys)
        return EmptyResult()

    def run_job_flow(self) -> ActionResult:
        instance_attrs = {
            "master_instance_type": self._get_param("Instances.MasterInstanceType"),
            "slave_instance_type": self._get_param("Instances.SlaveInstanceType"),
            "instance_count": self._get_int_param("Instances.InstanceCount", 1),
            "ec2_key_name": self._get_param("Instances.Ec2KeyName"),
            "ec2_subnet_id": self._get_param("Instances.Ec2SubnetId"),
            "hadoop_version": self._get_param("Instances.HadoopVersion"),
            "availability_zone": self._get_param(
                "Instances.Placement.AvailabilityZone", self.backend.region_name + "a"
            ),
            "keep_job_flow_alive_when_no_steps": self._get_bool_param(
                "Instances.KeepJobFlowAliveWhenNoSteps", False
            ),
            "termination_protected": self._get_bool_param(
                "Instances.TerminationProtected", False
            ),
            "emr_managed_master_security_group": self._get_param(
                "Instances.EmrManagedMasterSecurityGroup"
            ),
            "emr_managed_slave_security_group": self._get_param(
                "Instances.EmrManagedSlaveSecurityGroup"
            ),
            "service_access_security_group": self._get_param(
                "Instances.ServiceAccessSecurityGroup"
            ),
            "additional_master_security_groups": self._get_param(
                "Instances.AdditionalMasterSecurityGroups", []
            ),
            "additional_slave_security_groups": self._get_param(
                "Instances.AdditionalSlaveSecurityGroups", []
            ),
        }

        kwargs = {
            "name": self._get_param("Name"),
            "log_uri": self._get_param("LogUri"),
            "job_flow_role": self._get_param("JobFlowRole"),
            "service_role": self._get_param("ServiceRole"),
            "auto_scaling_role": self._get_param("AutoScalingRole"),
            "steps": self._get_param("Steps", []),
            "ebs_root_volume_iops": self._get_param("EbsRootVolumeIops"),
            "ebs_root_volume_size": self._get_param("EbsRootVolumeSize"),
            "ebs_root_volume_throughput": self._get_param("EbsRootVolumeThroughput"),
            "visible_to_all_users": self._get_bool_param("VisibleToAllUsers", False),
            "instance_attrs": instance_attrs,
        }

        bootstrap_actions = self._get_param("BootstrapActions", [])
        if bootstrap_actions:
            kwargs["bootstrap_actions"] = bootstrap_actions

        configurations = self._get_param("Configurations", [])
        if configurations:
            kwargs["configurations"] = configurations

        release_label = self._get_param("ReleaseLabel")
        ami_version = self._get_param("AmiVersion")
        if release_label:
            kwargs["release_label"] = release_label
            if ami_version:
                message = (
                    "Only one AMI version and release label may be specified. "
                    f"Provided AMI: {ami_version}, release label: {release_label}."
                )
                raise ValidationException(message)
        else:
            if ami_version:
                kwargs["requested_ami_version"] = ami_version
                kwargs["running_ami_version"] = ami_version
            else:
                kwargs["running_ami_version"] = "1.0.0"

        custom_ami_id = self._get_param("CustomAmiId")
        if custom_ami_id:
            kwargs["custom_ami_id"] = custom_ami_id
            if release_label and (
                ReleaseLabel(release_label) < ReleaseLabel("emr-5.7.0")
            ):
                message = "Custom AMI is not allowed"
                raise ValidationException(message)
            elif ami_version:
                message = "Custom AMI is not supported in this version of EMR"
                raise ValidationException(message)

        step_concurrency_level = self._get_param("StepConcurrencyLevel")
        if step_concurrency_level:
            kwargs["step_concurrency_level"] = step_concurrency_level

        security_configuration = self._get_param("SecurityConfiguration")
        if security_configuration:
            kwargs["security_configuration"] = security_configuration

        kerberos_attributes = self._get_param("KerberosAttributes", {})
        kwargs["kerberos_attributes"] = kerberos_attributes

        cluster = self.backend.run_job_flow(**kwargs)

        applications = self._get_param("Applications", [])
        if applications:
            self.backend.add_applications(cluster.id, applications)
        else:
            self.backend.add_applications(
                cluster.id, [{"Name": "Hadoop", "Version": "0.18"}]
            )

        instance_groups = self._get_param("Instances.InstanceGroups", [])
        if instance_groups:
            instance_group_result = self.backend.add_instance_groups(
                cluster.id, instance_groups
            )
            for i in range(0, len(instance_group_result)):
                self.backend.run_instances(
                    cluster.id, instance_groups[i], instance_group_result[i]
                )

        # TODO: Instances also must be created when `Instances.InstanceType` and `Instances.InstanceCount` are specified in the request.

        tags = self._get_param("Tags", [])
        if tags:
            self.backend.add_tags(cluster.id, {d["key"]: d["value"] for d in tags})
        result = {
            "JobFlowId": cluster.job_flow_id,
            "ClusterArn": cluster.arn,
        }
        return ActionResult(result)

    def set_termination_protection(self) -> ActionResult:
        termination_protection = self._get_bool_param("TerminationProtected")
        job_ids = self._get_param("JobFlowIds", [])
        self.backend.set_termination_protection(job_ids, termination_protection)
        return EmptyResult()

    def set_visible_to_all_users(self) -> ActionResult:
        visible_to_all_users = self._get_bool_param("VisibleToAllUsers", False)
        job_ids = self._get_param("JobFlowIds", [])
        self.backend.set_visible_to_all_users(job_ids, visible_to_all_users)
        return EmptyResult()

    def terminate_job_flows(self) -> ActionResult:
        job_ids = self._get_param("JobFlowIds")
        self.backend.terminate_job_flows(job_ids)
        return EmptyResult()

    def put_auto_scaling_policy(self) -> ActionResult:
        cluster_id = self._get_param("ClusterId")
        cluster = self.backend.describe_cluster(cluster_id)
        instance_group_id = self._get_param("InstanceGroupId")
        auto_scaling_policy = self._get_param("AutoScalingPolicy")
        instance_group = self.backend.put_auto_scaling_policy(
            instance_group_id, auto_scaling_policy
        )
        assert instance_group is not None
        result = {
            "ClusterId": cluster.id,
            "InstanceGroupId": instance_group.id,
            "AutoScalingPolicy": instance_group.auto_scaling_policy,
            "ClusterArn": cluster.arn,
        }
        return ActionResult(result)

    def remove_auto_scaling_policy(self) -> ActionResult:
        instance_group_id = self._get_param("InstanceGroupId")
        self.backend.remove_auto_scaling_policy(instance_group_id)
        return EmptyResult()

    def get_block_public_access_configuration(self) -> ActionResult:
        configuration = self.backend.get_block_public_access_configuration()
        config = configuration.get("block_public_access_configuration") or {}
        metadata = configuration.get("block_public_access_configuration_metadata") or {}
        result = {
            "BlockPublicAccessConfiguration": config,
            "BlockPublicAccessConfigurationMetadata": metadata,
        }
        return ActionResult(result)

    def put_block_public_access_configuration(self) -> ActionResult:
        params = self._get_params()
        block_public_access_configuration = (
            params.get("BlockPublicAccessConfiguration") or {}
        )
        self.backend.put_block_public_access_configuration(
            block_public_security_group_rules=block_public_access_configuration.get(
                "BlockPublicSecurityGroupRules"
            )
            or True,
            rule_ranges=block_public_access_configuration.get(
                "PermittedPublicSecurityGroupRuleRanges"
            ),
        )
        return EmptyResult()

    def list_release_labels(self) -> ActionResult:
        release_labels = self.backend.list_release_labels()
        return ActionResult({"ReleaseLabels": release_labels})

    def list_supported_instance_types(self) -> ActionResult:
        release_label = self._get_param("ReleaseLabel")
        instance_types = self.backend.list_supported_instance_types(release_label)
        return ActionResult({"SupportedInstanceTypes": instance_types})



from moto.core.responses import ActionResult, BaseResponse, EmptyResult

from .models import RedshiftBackend, redshift_backends


class RedshiftResponse(BaseResponse):
    RESPONSE_KEY_PATH_TO_TRANSFORMER = {
        "CreateClusterResult.Cluster.ClusterStatus": lambda _: "creating",
        "RestoreFromClusterSnapshotResult.Cluster.ClusterStatus": lambda _: "creating",
    }

    def __init__(self) -> None:
        super().__init__(service_name="redshift")
        self.automated_parameter_parsing = True

    @property
    def redshift_backend(self) -> RedshiftBackend:
        return redshift_backends[self.current_account][self.region]

    def create_cluster(self) -> ActionResult:
        cluster_kwargs = {
            "cluster_identifier": self._get_param("ClusterIdentifier"),
            "node_type": self._get_param("NodeType"),
            "master_username": self._get_param("MasterUsername"),
            "master_user_password": self._get_param("MasterUserPassword"),
            "db_name": self._get_param("DBName"),
            "cluster_type": self._get_param("ClusterType"),
            "cluster_security_groups": self._get_param("ClusterSecurityGroups", []),
            "vpc_security_group_ids": self._get_param("VpcSecurityGroupIds", []),
            "cluster_subnet_group_name": self._get_param("ClusterSubnetGroupName"),
            "availability_zone": self._get_param("AvailabilityZone"),
            "preferred_maintenance_window": self._get_param(
                "PreferredMaintenanceWindow"
            ),
            "cluster_parameter_group_name": self._get_param(
                "ClusterParameterGroupName"
            ),
            "automated_snapshot_retention_period": self._get_int_param(
                "AutomatedSnapshotRetentionPeriod"
            ),
            "port": self._get_int_param("Port"),
            "cluster_version": self._get_param("ClusterVersion"),
            "allow_version_upgrade": self._get_bool_param("AllowVersionUpgrade", True),
            "number_of_nodes": self._get_int_param("NumberOfNodes"),
            "publicly_accessible": self._get_bool_param("PubliclyAccessible", False),
            "encrypted": self._get_bool_param("Encrypted", False),
            "region_name": self.region,
            "tags": self._get_param("Tags", []),
            "iam_roles_arn": self._get_param("IamRoles", []),
            "enhanced_vpc_routing": self._get_bool_param("EnhancedVpcRouting", False),
            "kms_key_id": self._get_param("KmsKeyId"),
        }
        cluster = self.redshift_backend.create_cluster(**cluster_kwargs)
        return ActionResult({"Cluster": cluster})

    def pause_cluster(self) -> ActionResult:
        cluster_id = self._get_param("ClusterIdentifier")
        cluster = self.redshift_backend.pause_cluster(cluster_id)
        return ActionResult({"Cluster": cluster})

    def resume_cluster(self) -> ActionResult:
        cluster_id = self._get_param("ClusterIdentifier")
        cluster = self.redshift_backend.resume_cluster(cluster_id)
        return ActionResult({"Cluster": cluster})

    def restore_from_cluster_snapshot(self) -> ActionResult:
        enhanced_vpc_routing = self._get_bool_param("EnhancedVpcRouting")
        node_type = self._get_param("NodeType")
        number_of_nodes = self._get_int_param("NumberOfNodes")
        restore_kwargs = {
            "snapshot_identifier": self._get_param("SnapshotIdentifier"),
            "cluster_identifier": self._get_param("ClusterIdentifier"),
            "port": self._get_int_param("Port"),
            "availability_zone": self._get_param("AvailabilityZone"),
            "allow_version_upgrade": self._get_bool_param("AllowVersionUpgrade"),
            "cluster_subnet_group_name": self._get_param("ClusterSubnetGroupName"),
            "publicly_accessible": self._get_param("PubliclyAccessible"),
            "cluster_parameter_group_name": self._get_param(
                "ClusterParameterGroupName"
            ),
            "cluster_security_groups": self._get_param("ClusterSecurityGroups", []),
            "vpc_security_group_ids": self._get_param("VpcSecurityGroupIds", []),
            "preferred_maintenance_window": self._get_param(
                "PreferredMaintenanceWindow"
            ),
            "automated_snapshot_retention_period": self._get_int_param(
                "AutomatedSnapshotRetentionPeriod"
            ),
            "region_name": self.region,
            "iam_roles_arn": self._get_param("IamRoles", []),
        }
        if enhanced_vpc_routing is not None:
            restore_kwargs["enhanced_vpc_routing"] = enhanced_vpc_routing
        if node_type is not None:
            restore_kwargs["node_type"] = node_type
        if number_of_nodes is not None:
            restore_kwargs["number_of_nodes"] = number_of_nodes
        cluster = self.redshift_backend.restore_from_cluster_snapshot(**restore_kwargs)
        return ActionResult({"Cluster": cluster})

    def describe_clusters(self) -> ActionResult:
        cluster_identifier = self._get_param("ClusterIdentifier")
        tag_keys = self._get_param("TagKeys")
        clusters = self.redshift_backend.describe_clusters(cluster_identifier, tag_keys)

        return ActionResult({"Clusters": clusters})

    def modify_cluster(self) -> ActionResult:
        request_kwargs = {
            "cluster_identifier": self._get_param("ClusterIdentifier"),
            "new_cluster_identifier": self._get_param("NewClusterIdentifier"),
            "node_type": self._get_param("NodeType"),
            "master_user_password": self._get_param("MasterUserPassword"),
            "cluster_type": self._get_param("ClusterType"),
            "cluster_security_groups": self._get_param("ClusterSecurityGroups", []),
            "vpc_security_group_ids": self._get_param("VpcSecurityGroupIds", []),
            "cluster_subnet_group_name": self._get_param("ClusterSubnetGroupName"),
            "preferred_maintenance_window": self._get_param(
                "PreferredMaintenanceWindow"
            ),
            "cluster_parameter_group_name": self._get_param(
                "ClusterParameterGroupName"
            ),
            "automated_snapshot_retention_period": self._get_int_param(
                "AutomatedSnapshotRetentionPeriod"
            ),
            "cluster_version": self._get_param("ClusterVersion"),
            "allow_version_upgrade": self._get_bool_param("AllowVersionUpgrade"),
            "number_of_nodes": self._get_int_param("NumberOfNodes"),
            "publicly_accessible": self._get_param("PubliclyAccessible"),
            "encrypted": self._get_param("Encrypted"),
            "iam_roles_arn": self._get_param("IamRoles", []),
            "enhanced_vpc_routing": self._get_param("EnhancedVpcRouting"),
        }
        cluster_kwargs = {}
        # We only want parameters that were actually passed in, otherwise
        # we'll stomp all over our cluster metadata with None values.
        for key, value in request_kwargs.items():
            if value is not None and value != []:
                cluster_kwargs[key] = value

        cluster = self.redshift_backend.modify_cluster(**cluster_kwargs)

        return ActionResult({"Cluster": cluster})

    def delete_cluster(self) -> ActionResult:
        request_kwargs = {
            "cluster_identifier": self._get_param("ClusterIdentifier"),
            "final_cluster_snapshot_identifier": self._get_param(
                "FinalClusterSnapshotIdentifier"
            ),
            "skip_final_snapshot": self._get_bool_param("SkipFinalClusterSnapshot"),
        }

        cluster = self.redshift_backend.delete_cluster(**request_kwargs)

        return ActionResult({"Cluster": cluster})

    def create_cluster_subnet_group(self) -> ActionResult:
        cluster_subnet_group_name = self._get_param("ClusterSubnetGroupName")
        description = self._get_param("Description")
        subnet_ids = self._get_param("SubnetIds", [])
        tags = self._get_param("Tags", [])

        subnet_group = self.redshift_backend.create_cluster_subnet_group(
            cluster_subnet_group_name=cluster_subnet_group_name,
            description=description,
            subnet_ids=subnet_ids,
            region_name=self.region,
            tags=tags,
        )

        return ActionResult({"ClusterSubnetGroup": subnet_group})

    def describe_cluster_subnet_groups(self) -> ActionResult:
        subnet_identifier = self._get_param("ClusterSubnetGroupName")
        subnet_groups = self.redshift_backend.describe_cluster_subnet_groups(
            subnet_identifier
        )

        return ActionResult({"ClusterSubnetGroups": subnet_groups})

    def delete_cluster_subnet_group(self) -> ActionResult:
        subnet_identifier = self._get_param("ClusterSubnetGroupName")
        self.redshift_backend.delete_cluster_subnet_group(subnet_identifier)

        return EmptyResult()

    def create_cluster_security_group(self) -> ActionResult:
        cluster_security_group_name = self._get_param("ClusterSecurityGroupName")
        description = self._get_param("Description")
        tags = self._get_param("Tags", [])

        security_group = self.redshift_backend.create_cluster_security_group(
            cluster_security_group_name=cluster_security_group_name,
            description=description,
            tags=tags,
        )

        return ActionResult({"ClusterSecurityGroup": security_group})

    def describe_cluster_security_groups(self) -> ActionResult:
        cluster_security_group_name = self._get_param("ClusterSecurityGroupName")
        security_groups = self.redshift_backend.describe_cluster_security_groups(
            cluster_security_group_name
        )

        return ActionResult({"ClusterSecurityGroups": security_groups})

    def delete_cluster_security_group(self) -> ActionResult:
        security_group_identifier = self._get_param("ClusterSecurityGroupName")
        self.redshift_backend.delete_cluster_security_group(security_group_identifier)

        return EmptyResult()

    def authorize_cluster_security_group_ingress(self) -> ActionResult:
        cluster_security_group_name = self._get_param("ClusterSecurityGroupName")
        cidr_ip = self._get_param("CIDRIP")

        security_group = self.redshift_backend.authorize_cluster_security_group_ingress(
            cluster_security_group_name, cidr_ip
        )
        result = {
            "ClusterSecurityGroup": {
                "ClusterSecurityGroupName": cluster_security_group_name,
                "Description": security_group.description,
                "IPRanges": [
                    {
                        "Status": "authorized",
                        "CIDRIP": cidr_ip,
                        "Tags": security_group.tags,
                    },
                ],
            }
        }
        return ActionResult(result)

    def create_cluster_parameter_group(self) -> ActionResult:
        cluster_parameter_group_name = self._get_param("ParameterGroupName")
        group_family = self._get_param("ParameterGroupFamily")
        description = self._get_param("Description")
        tags = self._get_param("Tags", [])

        parameter_group = self.redshift_backend.create_cluster_parameter_group(
            cluster_parameter_group_name, group_family, description, self.region, tags
        )

        return ActionResult({"ClusterParameterGroup": parameter_group})

    def describe_cluster_parameter_groups(self) -> ActionResult:
        cluster_parameter_group_name = self._get_param("ParameterGroupName")
        parameter_groups = self.redshift_backend.describe_cluster_parameter_groups(
            cluster_parameter_group_name
        )

        return ActionResult({"ParameterGroups": parameter_groups})

    def delete_cluster_parameter_group(self) -> ActionResult:
        cluster_parameter_group_name = self._get_param("ParameterGroupName")
        self.redshift_backend.delete_cluster_parameter_group(
            cluster_parameter_group_name
        )

        return EmptyResult()

    def describe_default_cluster_parameters(self) -> ActionResult:
        family = self._get_param("ParameterGroupFamily")
        params = self.redshift_backend.describe_default_cluster_parameters()
        return ActionResult(
            {
                "DefaultClusterParameters": {
                    "ParameterGroupFamily": family,
                    "Parameters": params,
                }
            }
        )

    def describe_cluster_parameters(self) -> ActionResult:
        group_name = self._get_param("ParameterGroupName")
        params = self.redshift_backend.describe_cluster_parameters(
            parameter_group_name=group_name
        )
        return ActionResult(
            {
                "Parameters": params,
            }
        )

    def create_cluster_snapshot(self) -> ActionResult:
        cluster_identifier = self._get_param("ClusterIdentifier")
        snapshot_identifier = self._get_param("SnapshotIdentifier")
        tags = self._get_param("Tags", [])

        snapshot = self.redshift_backend.create_cluster_snapshot(
            cluster_identifier, snapshot_identifier, self.region, tags
        )
        return ActionResult({"Snapshot": snapshot})

    def describe_cluster_snapshots(self) -> ActionResult:
        cluster_identifier = self._get_param("ClusterIdentifier")
        snapshot_identifier = self._get_param("SnapshotIdentifier")
        snapshot_type = self._get_param("SnapshotType")
        snapshots = self.redshift_backend.describe_cluster_snapshots(
            cluster_identifier, snapshot_identifier, snapshot_type
        )
        return ActionResult({"Snapshots": snapshots})

    def delete_cluster_snapshot(self) -> ActionResult:
        snapshot_identifier = self._get_param("SnapshotIdentifier")
        snapshot = self.redshift_backend.delete_cluster_snapshot(snapshot_identifier)

        return ActionResult({"Snapshot": snapshot})

    def create_snapshot_copy_grant(self) -> ActionResult:
        copy_grant_kwargs = {
            "snapshot_copy_grant_name": self._get_param("SnapshotCopyGrantName"),
            "kms_key_id": self._get_param("KmsKeyId"),
            "region_name": self._get_param("Region"),
        }

        copy_grant = self.redshift_backend.create_snapshot_copy_grant(
            **copy_grant_kwargs
        )
        return ActionResult({"SnapshotCopyGrant": copy_grant})

    def delete_snapshot_copy_grant(self) -> ActionResult:
        copy_grant_kwargs = {
            "snapshot_copy_grant_name": self._get_param("SnapshotCopyGrantName")
        }
        self.redshift_backend.delete_snapshot_copy_grant(**copy_grant_kwargs)
        return EmptyResult()

    def describe_snapshot_copy_grants(self) -> ActionResult:
        copy_grant_kwargs = {
            "snapshot_copy_grant_name": self._get_param("SnapshotCopyGrantName")
        }

        copy_grants = self.redshift_backend.describe_snapshot_copy_grants(
            **copy_grant_kwargs
        )
        return ActionResult({"SnapshotCopyGrants": copy_grants})

    def create_tags(self) -> ActionResult:
        resource_name = self._get_param("ResourceName")
        tags = self._get_param("Tags", [])

        self.redshift_backend.create_tags(resource_name, tags)

        return EmptyResult()

    def describe_tags(self) -> ActionResult:
        resource_name = self._get_param("ResourceName")
        resource_type = self._get_param("ResourceType")

        tagged_resources = self.redshift_backend.describe_tags(
            resource_name, resource_type
        )
        return ActionResult({"TaggedResources": tagged_resources})

    def delete_tags(self) -> ActionResult:
        resource_name = self._get_param("ResourceName")
        tag_keys = self._get_param("TagKeys", [])

        self.redshift_backend.delete_tags(resource_name, tag_keys)

        return EmptyResult()

    def enable_snapshot_copy(self) -> ActionResult:
        snapshot_copy_kwargs = {
            "cluster_identifier": self._get_param("ClusterIdentifier"),
            "destination_region": self._get_param("DestinationRegion"),
            "retention_period": self._get_param("RetentionPeriod", 7),
            "snapshot_copy_grant_name": self._get_param("SnapshotCopyGrantName"),
        }
        cluster = self.redshift_backend.enable_snapshot_copy(**snapshot_copy_kwargs)

        return ActionResult({"Cluster": cluster})

    def disable_snapshot_copy(self) -> ActionResult:
        snapshot_copy_kwargs = {
            "cluster_identifier": self._get_param("ClusterIdentifier")
        }
        cluster = self.redshift_backend.disable_snapshot_copy(**snapshot_copy_kwargs)

        return ActionResult({"Cluster": cluster})

    def modify_snapshot_copy_retention_period(self) -> ActionResult:
        snapshot_copy_kwargs = {
            "cluster_identifier": self._get_param("ClusterIdentifier"),
            "retention_period": self._get_param("RetentionPeriod"),
        }
        cluster = self.redshift_backend.modify_snapshot_copy_retention_period(
            **snapshot_copy_kwargs
        )

        return ActionResult({"Clusters": [cluster]})

    def get_cluster_credentials(self) -> ActionResult:
        cluster_identifier = self._get_param("ClusterIdentifier")
        db_user = self._get_param("DbUser")
        auto_create = self._get_bool_param("AutoCreate", False)
        duration_seconds = self._get_int_param("DurationSeconds", 900)

        cluster_credentials = self.redshift_backend.get_cluster_credentials(
            cluster_identifier, db_user, auto_create, duration_seconds
        )

        return ActionResult(cluster_credentials)

    def enable_logging(self) -> ActionResult:
        cluster_identifier = self._get_param("ClusterIdentifier")
        bucket_name = self._get_param("BucketName")
        s3_key_prefix = self._get_param("S3KeyPrefix")
        log_destination_type = self._get_param("LogDestinationType")
        log_exports = self._get_param("LogExports")
        config = self.redshift_backend.enable_logging(
            cluster_identifier=cluster_identifier,
            bucket_name=bucket_name,
            s3_key_prefix=s3_key_prefix,
            log_destination_type=log_destination_type,
            log_exports=log_exports,
        )
        return ActionResult(config)

    def disable_logging(self) -> ActionResult:
        cluster_identifier = self._get_param("ClusterIdentifier")
        config = self.redshift_backend.disable_logging(
            cluster_identifier=cluster_identifier,
        )
        return ActionResult(config)

    def describe_logging_status(self) -> ActionResult:
        cluster_identifier = self._get_param("ClusterIdentifier")
        config = self.redshift_backend.describe_logging_status(
            cluster_identifier=cluster_identifier,
        )
        return ActionResult(config)



"""vpclattice base URL and path."""

from .responses import VPCLatticeResponse

url_bases = [
    r"https?://vpc-lattice\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/accesslogsubscriptions$": VPCLatticeResponse.dispatch,
    "{0}/accesslogsubscriptions/(?P<accessLogSubscriptionIdentifier>.+)$": VPCLatticeResponse.dispatch,
    "{0}/services$": VPCLatticeResponse.dispatch,
    "{0}/servicenetworks$": VPCLatticeResponse.dispatch,
    "{0}/servicenetworkvpcassociations$": VPCLatticeResponse.dispatch,
    "{0}/services/(?P<serviceIdentifier>[^/]+)/listeners/(?P<listenerIdentifier>[^/]+)/rules$": VPCLatticeResponse.dispatch,
    "{0}/tags/(?P<resourceArn>.+)$": VPCLatticeResponse.dispatch,
    "{0}/services/(?P<serviceIdentifier>.+)$": VPCLatticeResponse.dispatch,
    "{0}/servicenetworks/(?P<serviceNetworkIdentifier>.+)$": VPCLatticeResponse.dispatch,
    "{0}/tags/(?P<resourceArn>[^/]+)?tagKeys=(?P<tagKeys>[^/]+)$": VPCLatticeResponse.dispatch,
    "{0}/resourcepolicy/(?P<resourceIdentifier>.+)$": VPCLatticeResponse.dispatch,
    "{0}/authpolicy/(?P<resourceIdentifier>.+)$": VPCLatticeResponse.dispatch,
}



from typing import Any, Optional

from moto.core.responses import BaseResponse
from moto.core.serialize import return_if_not_empty

from ..exceptions import (
    DryRunClientError,
    EC2ClientError,
    EmptyTagSpecError,
    InvalidParameter,
)
from ..utils import convert_tag_spec


class EC2BaseResponse(BaseResponse):
    RESPONSE_KEY_PATH_TO_TRANSFORMER = {
        # ENI
        "DescribeNetworkInterfacesResult.NetworkInterfaces.NetworkInterface.Association": return_if_not_empty,
        # IAM Instance Profiles
        "AssociateIamInstanceProfileResult.IamInstanceProfileAssociation.State": lambda _: "associating",
        "DisassociateIamInstanceProfileResult.IamInstanceProfileAssociation.State": lambda _: "disassociating",
        "ReplaceIamInstanceProfileAssociationResult.IamInstanceProfileAssociation.State": lambda _: "associating",
        # Instances
        "Instances.Instance.Tags": return_if_not_empty,
        "RunInstances.Reservation.Instances.Instance.State": lambda _: {
            "Code": 0,
            "Name": "pending",
        },
        # Key Pairs
        "KeyPair.Tags": return_if_not_empty,
    }

    @property
    def ec2_backend(self) -> Any:  # type: ignore[misc]
        from moto.ec2.models import ec2_backends

        return ec2_backends[self.current_account][self.region]

    def _filters_from_querystring(self) -> dict[str, str]:
        # [{"Name": x1, "Value": y1}, ..]
        _filters = self._get_param("Filters", [])
        # EC2 is inconsistent with parameter name, so check singular as well.
        if not _filters:
            _filters = self._get_param("Filter", [])
        # return {x1: y1, ...}
        try:
            return {f["Name"]: f.get("Values", []) for f in _filters}
        except KeyError:
            raise EC2ClientError(
                "InvalidParameterValue", "The filter 'null' is invalid."
            )

    def _parse_tag_specification(
        self, expected_type: Optional[str] = None
    ) -> dict[str, dict[str, str]]:
        # [{"ResourceType": _type, "Tag": [{"Key": k, "Value": v}, ..]}]
        tag_spec_set = self._get_param("TagSpecification", [])
        if not tag_spec_set:
            tag_spec_set = self._get_param("TagSpecifications", [])
        if not tag_spec_set:
            return {}

        tags_dict = (
            tag_spec_set[0] if isinstance(tag_spec_set, list) else tag_spec_set
        )  # awscli allows for a json string to be passed and it should be allowed
        if "ResourceType" not in tags_dict:
            raise InvalidParameter("Tag specification resource type must have a value")
        if expected_type and tags_dict["ResourceType"] != expected_type:
            raise InvalidParameter(
                f"'{tags_dict['ResourceType']}' is not a valid taggable resource type for this operation."
            )
        if "Tags" not in tags_dict:
            if tags_dict.get("ResourceType") == "subnet":
                raise InvalidParameter("Tag specification must have at least one tag")
            raise EmptyTagSpecError

        return convert_tag_spec(tag_spec_set)

    def error_on_dryrun(self) -> None:
        if self._get_param("DryRun", False):
            raise DryRunClientError()



from typing import Optional

from moto.core.exceptions import ServiceException


class CloudFormationError(ServiceException):
    pass


class UnformattedGetAttTemplateException(Exception):
    description = (
        "Template error: resource {0} does not support attribute type {1} in Fn::GetAtt"
    )
    status_code = 400


class AlreadyExistsException(CloudFormationError):
    code = "AlreadyExistsException"


class ValidationError(CloudFormationError):
    code = "ValidationError"

    def __init__(self, name_or_id: Optional[str] = None, message: Optional[str] = None):
        # FIXME: The "stack does not exist" message should be provided by the caller, not here.
        if message is None:
            message = f"Stack with id {name_or_id} does not exist"
        super().__init__(message)


class MissingParameterError(CloudFormationError):
    code = "ValidationError"

    def __init__(self, parameter_name: str):
        message = f"Missing parameter {parameter_name}"
        super().__init__(message)


class ExportNotFound(CloudFormationError):
    code = "ValidationError"

    def __init__(self, export_name: str):
        message = f"No export named {export_name} found."
        super().__init__(message)


class StackSetNotEmpty(CloudFormationError):
    code = "StackSetNotEmptyException"
    message = "StackSet is not empty"


class StackSetNotFoundException(CloudFormationError):
    code = "StackSetNotFoundException"

    def __init__(self, name: str):
        message = f"StackSet {name} not found"
        super().__init__(message)


class UnsupportedAttribute(CloudFormationError):
    code = "ValidationError"

    def __init__(self, resource: str, attr: str):
        super().__init__(
            f"Template error: resource {resource} does not support attribute type {attr} in Fn::GetAtt"
        )



from datetime import datetime
from typing import Any, Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.utilities.paginator import paginate
from moto.utilities.tagging_service import TaggingService

from .data_models import (
    QuicksightAccountSettings,
    QuicksightDashboard,
    QuicksightDataSet,
    QuickSightDataSource,
    QuicksightGroup,
    QuicksightIngestion,
    QuicksightMembership,
    QuicksightUser,
)
from .exceptions import ResourceNotFoundException
from .utils import PAGINATION_MODEL, QuicksightSearchFilterFactory


def _create_id(aws_account_id: str, namespace: str, _id: str) -> str:
    return f"{aws_account_id}:{namespace}:{_id}"


class QuickSightBackend(BaseBackend):
    """Implementation of QuickSight APIs."""

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.dashboards: dict[str, QuicksightDashboard] = {}
        self.groups: dict[str, QuicksightGroup] = {}
        self.users: dict[str, QuicksightUser] = {}
        self.account_settings: QuicksightAccountSettings = QuicksightAccountSettings(
            account_id=account_id
        )
        self.data_sources: dict[str, QuickSightDataSource] = {}
        self.data_sets: dict[str, QuicksightDataSet] = {}
        self.tagger = TaggingService()

    def create_data_set(
        self,
        data_set_id: str,
        name: str,
        tags: Optional[list[dict[str, str]]] = None,
    ) -> QuicksightDataSet:
        dataset = QuicksightDataSet(
            self.account_id, self.region_name, data_set_id, name=name
        )

        if tags:
            self.tagger.tag_resource(
                arn=dataset.arn,
                tags=tags,
            )

        self.data_sets[data_set_id] = dataset
        return dataset

    def create_group(
        self, group_name: str, description: str, aws_account_id: str, namespace: str
    ) -> QuicksightGroup:
        group = QuicksightGroup(
            region=self.region_name,
            group_name=group_name,
            description=description,
            aws_account_id=aws_account_id,
            namespace=namespace,
        )
        _id = _create_id(aws_account_id, namespace, group_name)
        self.groups[_id] = group
        return group

    def create_group_membership(
        self, aws_account_id: str, namespace: str, group_name: str, member_name: str
    ) -> QuicksightMembership:
        group = self.describe_group(aws_account_id, namespace, group_name)
        return group.add_member(member_name)

    def create_ingestion(
        self, data_set_id: str, ingestion_id: str
    ) -> QuicksightIngestion:
        return QuicksightIngestion(
            self.account_id, self.region_name, data_set_id, ingestion_id
        )

    def delete_group(
        self, aws_account_id: str, namespace: str, group_name: str
    ) -> None:
        _id = _create_id(aws_account_id, namespace, group_name)
        self.groups.pop(_id, None)

    def delete_user(self, aws_account_id: str, namespace: str, user_name: str) -> None:
        # Delete users from all groups
        for group in self.groups.values():
            group.delete_member(user_name)
        # Delete user itself
        _id = _create_id(aws_account_id, namespace, user_name)
        self.users.pop(_id, None)

    def describe_group(
        self, aws_account_id: str, namespace: str, group_name: str
    ) -> QuicksightGroup:
        _id = _create_id(aws_account_id, namespace, group_name)
        if _id not in self.groups:
            raise ResourceNotFoundException(f"Group {group_name} not found")
        return self.groups[_id]

    def describe_group_membership(
        self, aws_account_id: str, namespace: str, group_name: str, member_name: str
    ) -> QuicksightMembership:
        group = self.describe_group(aws_account_id, namespace, group_name)
        member = group.get_member(member_name)
        if member is None:
            raise ResourceNotFoundException(f"Member {member_name} not found")
        return member

    def describe_user(
        self, aws_account_id: str, namespace: str, user_name: str
    ) -> QuicksightUser:
        _id = _create_id(aws_account_id, namespace, user_name)
        if _id not in self.users:
            raise ResourceNotFoundException(f"User {user_name} not found")
        return self.users[_id]

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_groups(self, aws_account_id: str, namespace: str) -> list[QuicksightGroup]:
        id_for_ns = _create_id(aws_account_id, namespace, _id="")
        return [
            group for _id, group in self.groups.items() if _id.startswith(id_for_ns)
        ]

    @paginate(pagination_model=PAGINATION_MODEL)
    def search_groups(
        self, aws_account_id: str, namespace: str, filters: list[dict[str, str]]
    ) -> list[QuicksightGroup]:
        id_for_ns = _create_id(aws_account_id, namespace, _id="")
        filter_list = QuicksightSearchFilterFactory.validate_and_create_filter(
            model_type=QuicksightGroup, input=filters
        )
        return [
            group
            for _id, group in self.groups.items()
            if _id.startswith(id_for_ns) and filter_list.match(group)
        ]

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_group_memberships(
        self, aws_account_id: str, namespace: str, group_name: str
    ) -> list[QuicksightMembership]:
        group = self.describe_group(aws_account_id, namespace, group_name)
        return group.list_members()

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_users(self, aws_account_id: str, namespace: str) -> list[QuicksightUser]:
        id_for_ns = _create_id(aws_account_id, namespace, _id="")
        return [user for _id, user in self.users.items() if _id.startswith(id_for_ns)]

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_user_groups(
        self, aws_account_id: str, namespace: str, user_name: str
    ) -> list[QuicksightGroup]:
        id_for_ns = _create_id(aws_account_id, namespace, _id="")
        group_list: dict[str, QuicksightGroup] = {}
        # Loop through all groups and check if the user is member.
        for id, group in self.groups.items():
            if group.get_member(user_name):
                group_list[id] = group
        return [group for _id, group in group_list.items() if _id.startswith(id_for_ns)]

    def register_user(
        self,
        identity_type: str,
        email: str,
        user_role: str,
        aws_account_id: str,
        namespace: str,
        user_name: str,
        tags: Optional[list[dict[str, str]]] = None,
    ) -> QuicksightUser:
        """
        The following parameters are not yet implemented:
        IamArn, SessionName, CustomsPermissionsName, ExternalLoginFederationProviderType, CustomFederationProviderUrl, ExternalLoginId
        """
        user = QuicksightUser(
            account_id=aws_account_id,
            region=self.region_name,
            email=email,
            identity_type=identity_type,
            user_role=user_role,
            username=user_name,
        )
        _id = _create_id(aws_account_id, namespace, user_name)

        if tags:
            self.tagger.tag_resource(
                arn=user.arn,
                tags=tags,
            )

        self.users[_id] = user
        return user

    def update_user(
        self,
        aws_account_id: str,
        namespace: str,
        user_name: str,
        email: str,
        user_role: str,
    ) -> QuicksightUser:
        user = self.describe_user(aws_account_id, namespace, user_name)
        user.email = email
        user.user_role = user_role
        return user

    def update_group(
        self, aws_account_id: str, namespace: str, group_name: str, description: str
    ) -> QuicksightGroup:
        group = self.describe_group(aws_account_id, namespace, group_name)
        group.description = description
        return group

    def create_dashboard(
        self,
        aws_account_id: str,
        dashboard_id: str,
        name: str,
        parameters: dict[str, Any],
        permissions: list[dict[str, Any]],
        source_entity: dict[str, Any],
        tags: list[dict[str, str]],
        version_description: str,
        dashboard_publish_options: dict[str, Any],
        theme_arn: str,
        definition: dict[str, Any],
        validation_strategy: dict[str, str],
        folder_arns: list[str],
        link_sharing_configuration: dict[str, Any],
        link_entities: list[str],
    ) -> QuicksightDashboard:
        dashboard = QuicksightDashboard(
            account_id=aws_account_id,
            region=self.region_name,
            dashboard_id=dashboard_id,
            dashboard_publish_options=dashboard_publish_options,
            name=name,
            definition=definition,
            folder_arns=folder_arns,
            link_entities=link_entities,
            link_sharing_configuration=link_sharing_configuration,
            parameters=parameters,
            permissions=permissions,
            source_entity=source_entity,
            tags=tags,
            theme_arn=theme_arn,
            version_description=version_description,
            validation_strategy=validation_strategy,
        )

        if tags:
            self.tagger.tag_resource(
                arn=dashboard.arn,
                tags=tags,
            )

        self.dashboards[dashboard_id] = dashboard
        return dashboard

    def describe_dashboard(
        self,
        aws_account_id: str,
        dashboard_id: str,
        version_number: int,
        alias_name: str,
    ) -> QuicksightDashboard:
        dashboard = self.dashboards.get(dashboard_id)
        if not dashboard:
            raise ResourceNotFoundException(f"Dashboard {dashboard_id} not found")
        return dashboard

    def list_dashboards(self, aws_account_id: str) -> list[dict[str, Any]]:
        dashboards = self.dashboards.values()
        dashboard_list: list[dict[str, Any]] = []
        for dashboard in dashboards:
            d_dict = {
                "Arn": dashboard.arn,
                "DashboardId": dashboard.dashboard_id,
                "Name": dashboard.name,
                "CreatedTime": str(dashboard.created_time),
                "LastUpdatedTime": str(dashboard.last_updated_time),
                "PublishedVersionNumber": dashboard.version_number,
                "LastPublishedTime": str(dashboard.last_published_time),
            }
            dashboard_list.append(d_dict)

        return dashboard_list

    def describe_account_settings(
        self, aws_account_id: str
    ) -> QuicksightAccountSettings:
        return self.account_settings

    def update_account_settings(
        self,
        aws_account_id: str,
        default_namespace: str,
        notification_email: str,
        termination_protection_enabled: bool,
    ) -> None:
        if notification_email:
            self.account_settings.notification_email = notification_email
        if termination_protection_enabled:
            self.account_settings.termination_protection_enabled = (
                termination_protection_enabled
            )

    def update_public_sharing_settings(
        self, aws_account_id: str, public_sharing_enabled: bool
    ) -> None:
        self.account_settings.public_sharing_enabled = public_sharing_enabled

    def create_data_source(
        self,
        aws_account_id: str,
        data_source_id: str,
        name: str,
        data_source_type: str,
        data_source_parameters: Optional[dict[str, dict[str, Any]]] = None,
        vpc_connection_properties: Optional[dict[str, Any]] = None,
        ssl_properties: Optional[dict[str, Any]] = None,
        tags: Optional[list[dict[str, str]]] = None,
    ) -> QuickSightDataSource:
        data_source = QuickSightDataSource(
            account_id=aws_account_id,
            region=self.region_name,
            data_source_id=data_source_id,
            name=name,
            status="CREATION_SUCCESSFUL",
            data_source_type=data_source_type,
            data_source_parameters=data_source_parameters,
            ssl_properties=ssl_properties,
            vpc_connection_properties=vpc_connection_properties,
            tags=tags,
        )

        if tags:
            self.tagger.tag_resource(
                arn=data_source.arn,
                tags=tags,
            )

        self.data_sources[data_source_id] = data_source
        return data_source

    def update_data_source(
        self,
        aws_account_id: str,
        data_source_id: str,
        name: str,
        data_source_parameters: Optional[dict[str, Any]] = None,
        vpc_connection_properties: Optional[dict[str, Any]] = None,
        ssl_properties: Optional[dict[str, Any]] = None,
    ) -> QuickSightDataSource:
        data_source = self.data_sources.get(data_source_id)

        if not data_source:
            raise ResourceNotFoundException(f"DataSource {data_source_id} Not Found")

        data_source.name = name
        data_source.data_source_parameters = data_source_parameters
        data_source.last_updated_time = datetime.now()
        data_source.status = "UPDATE_SUCCESSFUL"
        return data_source

    def delete_data_source(
        self, aws_account_id: str, data_source_id: str
    ) -> QuickSightDataSource:
        data_source = self.data_sources.pop(data_source_id, None)

        if data_source is None:
            raise ResourceNotFoundException(f"DataSource {data_source_id} Not Found")

        return data_source

    def describe_data_source(
        self, aws_account_id: str, data_source_id: str
    ) -> QuickSightDataSource:
        data_source = self.data_sources.get(data_source_id)

        if not data_source:
            raise ResourceNotFoundException(f"DataSource {data_source_id} Not Found")

        return data_source

    def list_data_sources(self, aws_account_id: str) -> list[dict[str, Any]]:
        data_sources = self.data_sources.values()
        data_source_list: list[dict[str, Any]] = []

        for data_source in data_sources:
            data_source_list.append(data_source.to_json())

        return data_source_list

    def tag_resource(self, resource_arn: str, tags: list[dict[str, str]]) -> None:
        self.tagger.tag_resource(
            arn=resource_arn,
            tags=tags,
        )

    def untag_resource(self, resource_arn: str, tag_keys: list[str]) -> None:
        self.tagger.untag_resource_using_names(
            resource_arn,
            tag_keys,
        )

    def list_tags_for_resource(self, arn: str) -> list[dict[str, str]]:
        tags = self.tagger.list_tags_for_resource(arn)
        return tags.get("Tags", [])


quicksight_backends = BackendDict(QuickSightBackend, "quicksight")



from moto.core.responses import ActionResult
from moto.ec2.utils import add_tag_specification

from ._base_response import EC2BaseResponse


class EgressOnlyInternetGateway(EC2BaseResponse):
    def create_egress_only_internet_gateway(self) -> ActionResult:
        vpc_id = self._get_param("VpcId")
        tag_param = self._get_param("TagSpecifications", [])
        tags = add_tag_specification(tag_param)

        egress_only_igw = self.ec2_backend.create_egress_only_internet_gateway(
            vpc_id=vpc_id, tags=tags
        )
        result = {"EgressOnlyInternetGateway": egress_only_igw}
        return ActionResult(result)

    def describe_egress_only_internet_gateways(self) -> ActionResult:
        egress_only_igw_ids = self._get_param("EgressOnlyInternetGatewayIds", [])
        egress_only_igws = self.ec2_backend.describe_egress_only_internet_gateways(
            egress_only_igw_ids
        )
        result = {"EgressOnlyInternetGateways": egress_only_igws}
        return ActionResult(result)

    def delete_egress_only_internet_gateway(self) -> ActionResult:
        egress_only_igw_id = self._get_param("EgressOnlyInternetGatewayId")
        self.ec2_backend.delete_egress_only_internet_gateway(
            gateway_id=egress_only_igw_id
        )
        return ActionResult({"ReturnCode": True})



from functools import partial

from moto.utilities.utils import get_partition


def make_arn(
    name: str, account_id: str, region_name: str, _id: str, scope: str, resource: str
) -> str:
    if scope == "REGIONAL":
        scope = "regional"
    elif scope == "CLOUDFRONT":
        scope = "global"
        # cloudfront global scope is managed from us-east-1 region: https://docs.aws.amazon.com/waf/latest/APIReference/API_CreateIPSet.html#WAF-CreateIPSet-request-Scope
        # base_backend stores global region as "aws": https://github.com/getmoto/moto/blob/d00aa025b6c3d37977508b5d5e81ecad4ca15159/moto/core/base_backend.py#L272
        # Therefore needs to be overriden here to correctly form the ARN
        region_name = "us-east-1"

    partition = get_partition(region_name)

    return f"arn:{partition}:wafv2:{region_name}:{account_id}:{scope}/{resource}/{name}/{_id}"


make_arn_for_wacl = partial(make_arn, resource="webacl")
make_arn_for_ip_set = partial(make_arn, resource="ipset")
make_arn_for_logging_configuration = partial(make_arn, resource="loggingconfiguration")
make_arn_for_rule_group = partial(make_arn, resource="rulegroup")


def make_arn_for_regex_pattern_set(
    name: str, account_id: str, region_name: str, _id: str, scope: str
) -> str:
    """
    Generate an ARN for a WAFv2 Regex Pattern Set
    """
    if scope == "CLOUDFRONT":
        region_name = "us-east-1"
    return f"arn:aws:wafv2:{region_name}:{account_id}:regional/regexpatternset/{name}/{_id}"



"""SESV2Backend class with methods for supported APIs."""

from typing import Any, Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.utilities.paginator import paginate

from ..ses.exceptions import NotFoundException
from ..ses.models import (
    ConfigurationSet,
    Contact,
    ContactList,
    DedicatedIpPool,
    EmailIdentity,
    Message,
    RawMessage,
    ses_backends,
)
from ..ses.utils import get_arn

PAGINATION_MODEL = {
    "list_dedicated_ip_pools": {
        "input_token": "next_token",
        "limit_key": "page_size",
        "limit_default": 100,
        "unique_attribute": ["pool_name"],
    },
    "list_email_identities": {
        "input_token": "next_token",
        "limit_key": "page_size",
        "limit_default": 100,
        "unique_attribute": "IdentityName",
    },
    "list_configuration_sets": {
        "input_token": "next_token",
        "limit_key": "page_size",
        "limit_default": 100,
        "unique_attribute": "configuration_set_name",
    },
}


# TODO
# ListTagsForResource, TagResource, UntagResource to do
class SESV2Backend(BaseBackend):
    """Implementation of SESV2 APIs, piggy back on v1 SES"""

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)

        # Store local variables in v1 backend for interoperability
        self.core_backend = ses_backends[self.account_id][self.region_name]

    def create_contact_list(self, params: dict[str, Any]) -> None:
        name = params["ContactListName"]
        description = params.get("Description")
        topics = [] if "Topics" not in params else params["Topics"]
        new_list = ContactList(name, str(description), topics)

        if params.get("Tags"):
            self.core_backend.tagger.tag_resource(
                arn=get_arn(self, "ses", f"contact-list/{name}"),
                tags=params["Tags"],
            )

        self.core_backend.contacts_lists[name] = new_list

    def get_contact_list(self, contact_list_name: str) -> ContactList:
        if contact_list_name in self.core_backend.contacts_lists:
            return self.core_backend.contacts_lists[contact_list_name]
        else:
            raise NotFoundException(
                f"List with name: {contact_list_name} doesn't exist."
            )

    def list_contact_lists(self) -> list[ContactList]:
        return self.core_backend.contacts_lists.values()  # type: ignore[return-value]

    def delete_contact_list(self, name: str) -> None:
        if name in self.core_backend.contacts_lists:
            del self.core_backend.contacts_lists[name]
        else:
            raise NotFoundException(f"List with name: {name} doesn't exist")

    def create_contact(self, contact_list_name: str, params: dict[str, Any]) -> None:
        contact_list = self.get_contact_list(contact_list_name)
        contact_list.create_contact(contact_list_name, params)
        return

    def get_contact(self, email: str, contact_list_name: str) -> Contact:
        contact_list = self.get_contact_list(contact_list_name)
        contact = contact_list.get_contact(email)
        return contact

    def list_contacts(self, contact_list_name: str) -> list[Contact]:
        contact_list = self.get_contact_list(contact_list_name)
        contacts = contact_list.list_contacts()
        return contacts

    def delete_contact(self, email: str, contact_list_name: str) -> None:
        contact_list = self.get_contact_list(contact_list_name)
        contact_list.delete_contact(email)
        return

    def send_email(
        self, source: str, destinations: dict[str, list[str]], subject: str, body: str
    ) -> Message:
        message = self.core_backend.send_email(
            source=source,
            destinations=destinations,
            subject=subject,
            body=body,
        )
        return message

    def send_raw_email(
        self, source: str, destinations: list[str], raw_data: str
    ) -> RawMessage:
        message = self.core_backend.send_raw_email(
            source=source, destinations=destinations, raw_data=raw_data
        )
        return message

    def create_email_identity(
        self,
        email_identity: str,
        tags: Optional[list[dict[str, str]]],
        dkim_signing_attributes: Optional[object],
        configuration_set_name: Optional[str],
    ) -> EmailIdentity:
        if tags:
            self.core_backend.tagger.tag_resource(
                arn=get_arn(self, "ses", f"email-identity/{email_identity}"),
                tags=tags,
            )

        return self.core_backend.create_email_identity_v2(
            email_identity, tags, dkim_signing_attributes, configuration_set_name
        )

    def delete_email_identity(
        self,
        email_identity: str,
    ) -> None:
        self.core_backend.delete_identity(email_identity)

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_email_identities(self) -> list[EmailIdentity]:
        identities = list(self.core_backend.email_identities.values())
        return identities

    def create_configuration_set(
        self,
        configuration_set_name: str,
        tracking_options: dict[str, str],
        delivery_options: dict[str, Any],
        reputation_options: dict[str, Any],
        sending_options: dict[str, bool],
        tags: list[dict[str, str]],
        suppression_options: dict[str, list[str]],
        vdm_options: dict[str, dict[str, str]],
    ) -> None:
        self.core_backend.create_configuration_set_v2(
            configuration_set_name=configuration_set_name,
            tracking_options=tracking_options,
            delivery_options=delivery_options,
            reputation_options=reputation_options,
            sending_options=sending_options,
            tags=tags,
            suppression_options=suppression_options,
            vdm_options=vdm_options,
        )

    def delete_configuration_set(self, configuration_set_name: str) -> None:
        self.core_backend.delete_configuration_set(configuration_set_name)

    def get_configuration_set(self, configuration_set_name: str) -> ConfigurationSet:
        config_set = self.core_backend.describe_configuration_set(
            configuration_set_name=configuration_set_name
        )
        return config_set

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_configuration_sets(self) -> list[ConfigurationSet]:
        return self.core_backend._list_all_configuration_sets()

    def create_dedicated_ip_pool(
        self, pool_name: str, tags: list[dict[str, str]], scaling_mode: str
    ) -> None:
        if pool_name not in self.core_backend.dedicated_ip_pools:
            new_pool = DedicatedIpPool(
                pool_name=pool_name, tags=tags, scaling_mode=scaling_mode
            )

            if tags:
                self.core_backend.tagger.tag_resource(
                    arn=get_arn(self, "ses", f"dedicated-ip-pool/{pool_name}"),
                    tags=tags,
                )

            self.core_backend.dedicated_ip_pools[pool_name] = new_pool

    def delete_dedicated_ip_pool(self, pool_name: str) -> None:
        self.core_backend.dedicated_ip_pools.pop(pool_name)

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_dedicated_ip_pools(self) -> list[str]:
        return list(self.core_backend.dedicated_ip_pools.keys())

    def get_dedicated_ip_pool(self, pool_name: str) -> DedicatedIpPool:
        if not self.core_backend.dedicated_ip_pools.get(pool_name, None):
            raise NotFoundException(pool_name)
        return self.core_backend.dedicated_ip_pools[pool_name]

    def get_email_identity(self, email_identity: str) -> EmailIdentity:
        if email_identity not in self.core_backend.email_identities:
            raise NotFoundException(email_identity)
        return self.core_backend.email_identities[email_identity]

    def create_email_identity_policy(
        self, email_identity: str, policy_name: str, policy: str
    ) -> None:
        email_id = self.get_email_identity(email_identity)

        email_id.policies[policy_name] = policy

        return

    def delete_email_identity_policy(
        self, email_identity: str, policy_name: str
    ) -> None:
        if email_identity not in self.core_backend.email_identities:
            raise NotFoundException(email_identity)

        email_id = self.core_backend.email_identities[email_identity]

        if policy_name in email_id.policies:
            del email_id.policies[policy_name]

        return

    def update_email_identity_policy(
        self, email_identity: str, policy_name: str, policy: str
    ) -> None:
        if email_identity not in self.core_backend.email_identities:
            raise NotFoundException(email_identity)

        email_id = self.core_backend.email_identities[email_identity]

        email_id.policies[policy_name] = policy

        return

    def get_email_identity_policies(self, email_identity: str) -> dict[str, Any]:
        email_id = self.get_email_identity(email_identity)

        return email_id.policies

    def tag_resource(self, resource_arn: str, tags: list[dict[str, str]]) -> None:
        self.core_backend.tagger.tag_resource(resource_arn, tags)

    def untag_resource(self, resource_arn: str, tag_keys: list[str]) -> None:
        self.core_backend.tagger.untag_resource_using_names(resource_arn, tag_keys)

    def list_tags_for_resource(self, resource_arn: str) -> list[dict[str, str]]:
        tags = self.core_backend.tagger.list_tags_for_resource(resource_arn)
        return tags.get("Tags", [])


sesv2_backends = BackendDict(SESV2Backend, "sesv2")



from typing import Final

from moto.stepfunctions.parser.asl.component.common.string.string_expression import (
    StringSampler,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment


class NoSuchVariable:
    def __init__(self, path: str):
        self.path: Final[str] = path


class Variable(EvalComponent):
    string_sampler: Final[StringSampler]

    def __init__(self, string_sampler: StringSampler):
        self.string_sampler = string_sampler

    def _eval_body(self, env: Environment) -> None:
        try:
            self.string_sampler.eval(env=env)
            value = env.stack.pop()
        except Exception as ex:
            value = NoSuchVariable(f"{self.string_sampler.literal_value}, {ex}")
        env.stack.append(value)



"""ConnectBackend class with methods for supported APIs."""

import uuid
from datetime import datetime, timezone
from typing import Any, Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.utilities.paginator import paginate
from moto.utilities.tagging_service import TaggingService

from .exceptions import InvalidParameterException, ResourceNotFoundException

PAGINATION_MODEL = {
    "list_analytics_data_associations": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 1000,
        "unique_attribute": "DataSetId",
        "output_token": "next_token",
    },
    "list_instances": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 10,
        "unique_attribute": "Id",
        "output_token": "next_token",
    },
}


class Instance(BaseModel):
    def __init__(
        self,
        identity_management_type: str,
        inbound_calls_enabled: bool,
        outbound_calls_enabled: bool,
        account_id: str,
        region: str,
        instance_alias: Optional[str] = None,
        tags: Optional[dict[str, str]] = None,
    ) -> None:
        self.id = str(uuid.uuid4())
        self.arn = f"arn:aws:connect:{region}:{account_id}:instance/{self.id}"
        self.identity_management_type = identity_management_type
        self.instance_alias = instance_alias
        self.created_time = datetime.now(timezone.utc)
        self.service_role = f"arn:aws:iam::{account_id}:role/aws-service-role/connect.amazonaws.com/AWSServiceRoleForAmazonConnect"
        self.instance_status = "ACTIVE"
        self.inbound_calls_enabled = inbound_calls_enabled
        self.outbound_calls_enabled = outbound_calls_enabled
        self.instance_access_url = f"https://{instance_alias or self.id}.my.connect.aws"
        self.tags = tags or {}

    def to_dict(self) -> dict[str, Any]:
        """Return full instance details for DescribeInstance."""
        result: dict[str, Any] = {
            "Id": self.id,
            "Arn": self.arn,
            "IdentityManagementType": self.identity_management_type,
            "CreatedTime": self.created_time.isoformat(),
            "ServiceRole": self.service_role,
            "InstanceStatus": self.instance_status,
            "InboundCallsEnabled": self.inbound_calls_enabled,
            "OutboundCallsEnabled": self.outbound_calls_enabled,
            "InstanceAccessUrl": self.instance_access_url,
        }
        if self.instance_alias:
            result["InstanceAlias"] = self.instance_alias
        if self.tags:
            result["Tags"] = self.tags
        return result

    def to_summary_dict(self) -> dict[str, Any]:
        """Return summary for ListInstances."""
        result: dict[str, Any] = {
            "Id": self.id,
            "Arn": self.arn,
            "IdentityManagementType": self.identity_management_type,
            "CreatedTime": self.created_time.isoformat(),
            "ServiceRole": self.service_role,
            "InstanceStatus": self.instance_status,
            "InboundCallsEnabled": self.inbound_calls_enabled,
            "OutboundCallsEnabled": self.outbound_calls_enabled,
            "InstanceAccessUrl": self.instance_access_url,
        }
        if self.instance_alias:
            result["InstanceAlias"] = self.instance_alias
        return result


class AnalyticsDataAssociation(BaseModel):
    """Represents an analytics data association for a Connect instance.

    Storage model: One association per (instance_id, data_set_id) pair.
    A data_set_id can only be associated once per instance.
    The target_account_id specifies where the data is shared to.
    """

    def __init__(
        self,
        instance_id: str,
        data_set_id: str,
        target_account_id: str,
        source_account_id: str,
        region: str,
    ) -> None:
        self.instance_id = instance_id
        self.data_set_id = data_set_id
        self.target_account_id = target_account_id
        self.resource_share_id = str(uuid.uuid4())
        self.resource_share_arn = f"arn:aws:ram:{region}:{source_account_id}:resource-share/{self.resource_share_id}"
        self.resource_share_status = "ACTIVE"

    def to_dict(self) -> dict[str, str]:
        return {
            "DataSetId": self.data_set_id,
            "TargetAccountId": self.target_account_id,
            "ResourceShareId": self.resource_share_id,
            "ResourceShareArn": self.resource_share_arn,
            "ResourceShareStatus": self.resource_share_status,
        }


class ConnectBackend(BaseBackend):
    """Backend for Amazon Connect API."""

    def __init__(self, region_name: str, account_id: str) -> None:
        super().__init__(region_name, account_id)
        self.instances: dict[str, Instance] = {}
        self.analytics_data_associations: dict[
            str, dict[str, AnalyticsDataAssociation]
        ] = {}
        self.tagger = TaggingService()

    def tag_resource(self, resource_arn: str, tags: dict[str, str]) -> None:
        self.tagger.tag_resource(
            resource_arn, TaggingService.convert_dict_to_tags_input(tags)
        )

    def untag_resource(self, resource_arn: str, tag_keys: list[str]) -> None:
        self.tagger.untag_resource_using_names(resource_arn, tag_keys)

    def list_tags_for_resource(self, resource_arn: str) -> dict[str, str]:
        return self.tagger.get_tag_dict_for_resource(resource_arn)

    def _get_instance_or_raise(self, instance_id: str) -> Instance:
        """Get instance by ID or raise ResourceNotFoundException."""
        if instance_id not in self.instances:
            raise ResourceNotFoundException(f"Instance with id {instance_id} not found")
        return self.instances[instance_id]

    def create_instance(
        self,
        identity_management_type: str,
        inbound_calls_enabled: bool,
        outbound_calls_enabled: bool,
        instance_alias: Optional[str] = None,
        tags: Optional[dict[str, str]] = None,
    ) -> dict[str, str]:
        """Create a new Connect instance."""
        if not identity_management_type:
            raise InvalidParameterException(
                "IdentityManagementType is a required parameter"
            )

        instance = Instance(
            identity_management_type=identity_management_type,
            inbound_calls_enabled=inbound_calls_enabled,
            outbound_calls_enabled=outbound_calls_enabled,
            instance_alias=instance_alias,
            account_id=self.account_id,
            region=self.region_name,
            tags=tags,
        )
        self.instances[instance.id] = instance

        if tags:
            self.tag_resource(instance.arn, tags)

        return {
            "Id": instance.id,
            "Arn": instance.arn,
        }

    def describe_instance(self, instance_id: str) -> dict[str, Any]:
        """Describe a Connect instance."""
        if not instance_id:
            raise InvalidParameterException("InstanceId is a required parameter")

        instance = self._get_instance_or_raise(instance_id)
        return instance.to_dict()

    @paginate(pagination_model=PAGINATION_MODEL)  # type: ignore[misc]
    def list_instances(
        self,
        max_results: Optional[int] = None,
        next_token: Optional[str] = None,
    ) -> list[dict[str, str]]:
        sorted_instances = sorted(
            self.instances.values(),
            key=lambda i: (i.created_time, i.id),
        )
        return [instance.to_summary_dict() for instance in sorted_instances]

    def delete_instance(self, instance_id: str) -> None:
        if not instance_id:
            raise InvalidParameterException("InstanceId is a required parameter")

        self._get_instance_or_raise(instance_id)

        del self.instances[instance_id]
        # Also clean up any analytics associations for this instance
        if instance_id in self.analytics_data_associations:
            del self.analytics_data_associations[instance_id]

    def associate_analytics_data_set(
        self,
        instance_id: str,
        data_set_id: str,
        target_account_id: Optional[str] = None,
    ) -> dict[str, str]:
        """Associate an analytics data set with a Connect instance.

        Enforces one association per (instance_id, data_set_id) pair.
        """
        if not instance_id:
            raise InvalidParameterException("InstanceId is a required parameter")
        if not data_set_id:
            raise InvalidParameterException("DataSetId is a required parameter")

        self._get_instance_or_raise(instance_id)

        if not target_account_id:
            target_account_id = self.account_id

        if instance_id in self.analytics_data_associations:
            if data_set_id in self.analytics_data_associations[instance_id]:
                raise InvalidParameterException(
                    f"Analytics data association for data set {data_set_id} already exists for instance {instance_id}"
                )

        if instance_id not in self.analytics_data_associations:
            self.analytics_data_associations[instance_id] = {}

        association = AnalyticsDataAssociation(
            instance_id=instance_id,
            data_set_id=data_set_id,
            target_account_id=target_account_id,
            source_account_id=self.account_id,
            region=self.region_name,
        )
        self.analytics_data_associations[instance_id][data_set_id] = association

        return {
            "DataSetId": association.data_set_id,
            "TargetAccountId": association.target_account_id,
            "ResourceShareId": association.resource_share_id,
            "ResourceShareArn": association.resource_share_arn,
            "ResourceShareStatus": association.resource_share_status,
        }

    def disassociate_analytics_data_set(
        self,
        instance_id: str,
        data_set_id: str,
    ) -> None:
        """Disassociate an analytics data set from a Connect instance."""
        if not instance_id:
            raise InvalidParameterException("InstanceId is a required parameter")
        if not data_set_id:
            raise InvalidParameterException("DataSetId is a required parameter")

        self._get_instance_or_raise(instance_id)

        if instance_id not in self.analytics_data_associations:
            raise ResourceNotFoundException(
                f"Analytics data association for data set {data_set_id} not found for instance {instance_id}"
            )

        if data_set_id not in self.analytics_data_associations[instance_id]:
            raise ResourceNotFoundException(
                f"Analytics data association for data set {data_set_id} not found for instance {instance_id}"
            )

        del self.analytics_data_associations[instance_id][data_set_id]

        if not self.analytics_data_associations[instance_id]:
            del self.analytics_data_associations[instance_id]

    @paginate(pagination_model=PAGINATION_MODEL)  # type: ignore[misc]
    def list_analytics_data_associations(
        self,
        instance_id: str,
        data_set_id: Optional[str] = None,
        max_results: Optional[int] = None,
        next_token: Optional[str] = None,
    ) -> list[dict[str, str]]:
        """List analytics data associations for a Connect instance."""
        if not instance_id:
            raise InvalidParameterException("InstanceId is a required parameter")

        self._get_instance_or_raise(instance_id)

        if instance_id not in self.analytics_data_associations:
            return []

        associations = list(self.analytics_data_associations[instance_id].values())

        if data_set_id:
            associations = [a for a in associations if a.data_set_id == data_set_id]

        associations.sort(key=lambda a: a.data_set_id)

        return [a.to_dict() for a in associations]


connect_backends = BackendDict(ConnectBackend, "connect")



import abc
from typing import Final

from moto.stepfunctions.parser.api import ExecutionFailedEventDetails, HistoryEventType
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.common.string.string_expression import (
    StringJSONata,
    StringSampler,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str

DEFAULT_MAX_CONCURRENCY_VALUE: Final[int] = 0  # No limit.


class MaxConcurrencyDecl(EvalComponent, abc.ABC):
    @abc.abstractmethod
    def _eval_max_concurrency(self, env: Environment) -> int: ...

    def _eval_body(self, env: Environment) -> None:
        max_concurrency_value = self._eval_max_concurrency(env=env)
        env.stack.append(max_concurrency_value)


class MaxConcurrency(MaxConcurrencyDecl):
    max_concurrency_value: Final[int]

    def __init__(self, num: int = DEFAULT_MAX_CONCURRENCY_VALUE):
        super().__init__()
        self.max_concurrency_value = num

    def _eval_max_concurrency(self, env: Environment) -> int:
        return self.max_concurrency_value


class MaxConcurrencyJSONata(MaxConcurrencyDecl):
    string_jsonata: Final[StringJSONata]

    def __init__(self, string_jsonata: StringJSONata):
        super().__init__()
        self.string_jsonata = string_jsonata

    def _eval_max_concurrency(self, env: Environment) -> int:
        self.string_jsonata.eval(env=env)
        # TODO: add snapshot tests to verify AWS's behaviour about non integer values.
        seconds = int(env.stack.pop())
        return seconds


class MaxConcurrencyPath(MaxConcurrency):
    string_sampler: Final[StringSampler]

    def __init__(self, string_sampler: StringSampler):
        super().__init__()
        self.string_sampler = string_sampler

    def _eval_max_concurrency(self, env: Environment) -> int:
        self.string_sampler.eval(env=env)
        max_concurrency_value = env.stack.pop()

        if not isinstance(max_concurrency_value, int):
            try:
                max_concurrency_value = int(max_concurrency_value)
            except Exception:
                # Pass the wrong type forward.
                pass

        error_cause = None
        if not isinstance(max_concurrency_value, int):
            value_str = (
                to_json_str(max_concurrency_value)
                if not isinstance(max_concurrency_value, str)
                else max_concurrency_value
            )
            error_cause = f'The MaxConcurrencyPath field refers to value "{value_str}" which is not a valid integer: {self.string_sampler.literal_value}'
        elif max_concurrency_value < 0:
            error_cause = f"Expected non-negative integer for MaxConcurrency, got '{max_concurrency_value}' instead."

        if error_cause is not None:
            raise FailureEventException(
                failure_event=FailureEvent(
                    env=env,
                    error_name=StatesErrorName(typ=StatesErrorNameType.StatesRuntime),
                    event_type=HistoryEventType.ExecutionFailed,
                    event_details=EventDetails(
                        executionFailedEventDetails=ExecutionFailedEventDetails(
                            error=StatesErrorNameType.StatesRuntime.to_name(),
                            cause=error_cause,
                        )
                    ),
                )
            )

        return max_concurrency_value



import re
from typing import Any, Optional
from urllib.parse import urlparse

from moto.core.responses import ActionResult, BaseResponse, EmptyResult

from .constants import (
    DEFAULT_RECEIVED_MESSAGES,
    MAXIMUM_VISIBILITY_TIMEOUT,
)
from .exceptions import (
    BatchEntryIdsNotDistinct,
    EmptyBatchRequest,
    InvalidAttributeName,
    MaxVisibilityTimeout,
    QueueDoesNotExist,
    SQSException,
)
from .models import SQSBackend, sqs_backends
from .utils import validate_message_attributes


class SQSResponse(BaseResponse):
    region_regex = re.compile(r"://(.+?)\.queue\.amazonaws\.com")

    def __init__(self) -> None:
        super().__init__(service_name="sqs")
        self.automated_parameter_parsing = True

    @property
    def sqs_backend(self) -> SQSBackend:
        return sqs_backends[self.current_account][self.region]

    def _determine_resource(self) -> str:
        queue_name = self._get_queue_name()
        try:
            return self.sqs_backend.get_queue(queue_name).queue_arn
        except QueueDoesNotExist:
            return "*"

    def _get_queue_name(self) -> str:
        try:
            queue_url = self._get_param("QueueUrl")
            if queue_url.startswith("http://") or queue_url.startswith("https://"):
                return queue_url.split("/")[-1]
            else:
                # The parameter could be the name itself, which AWS also accepts
                return queue_url
        except (AttributeError, TypeError):
            pass
        # Fallback to reading from the URL for botocore
        return self.path.split("/")[-1]

    def _get_validated_visibility_timeout(self, timeout: Optional[str] = None) -> int:
        """
        :raises ValueError: If specified visibility timeout exceeds MAXIMUM_VISIBILITY_TIMEOUT
        :raises TypeError: If visibility timeout was not specified
        """
        if timeout is not None:
            visibility_timeout = int(timeout)
        else:
            visibility_timeout = self._get_param("VisibilityTimeout")
        if visibility_timeout > MAXIMUM_VISIBILITY_TIMEOUT:
            raise ValueError
        return visibility_timeout

    def create_queue(self) -> ActionResult:
        request_url = urlparse(self.uri)
        queue_name = self._get_param("QueueName")
        attributes = self._get_param("Attributes", {})
        tags = self._get_param("tags", {})
        queue = self.sqs_backend.create_queue(queue_name, tags, **attributes)
        result = {"QueueUrl": queue.url(request_url)}
        return ActionResult(result)

    def get_queue_url(self) -> ActionResult:
        request_url = urlparse(self.uri)
        queue_name = self._get_param("QueueName")
        queue = self.sqs_backend.get_queue_url(queue_name)
        result = {"QueueUrl": queue.url(request_url)}
        return ActionResult(result)

    def list_queues(self) -> ActionResult:
        request_url = urlparse(self.uri)
        queue_name_prefix = self._get_param("QueueNamePrefix")
        queues = self.sqs_backend.list_queues(queue_name_prefix)
        result = {}
        if queues:
            result["QueueUrls"] = [queue.url(request_url) for queue in queues]
        return ActionResult(result)

    def change_message_visibility(self) -> ActionResult:
        queue_name = self._get_queue_name()
        receipt_handle = self._get_param("ReceiptHandle")
        try:
            visibility_timeout = self._get_validated_visibility_timeout()
        except ValueError:
            raise MaxVisibilityTimeout()
        self.sqs_backend.change_message_visibility(
            queue_name=queue_name,
            receipt_handle=receipt_handle,
            visibility_timeout=visibility_timeout,
        )
        return EmptyResult()

    def change_message_visibility_batch(self) -> ActionResult:
        queue_name = self._get_queue_name()
        entries = self._get_param("Entries", [])
        success, failed = self.sqs_backend.change_message_visibility_batch(
            queue_name, entries
        )
        result = {"Successful": [{"Id": _id} for _id in success], "Failed": failed}
        return ActionResult(result)

    def get_queue_attributes(self) -> ActionResult:
        queue_name = self._get_queue_name()
        attribute_names = self._get_param("AttributeNames", [])
        if attribute_names and "" in attribute_names:
            raise InvalidAttributeName("")
        attributes = self.sqs_backend.get_queue_attributes(queue_name, attribute_names)
        result = {
            "Attributes": {
                key: str(value)
                for key, value in attributes.items()
                if value is not None
            }
        }
        if not result["Attributes"]:
            return EmptyResult()
        return ActionResult(result)

    def set_queue_attributes(self) -> ActionResult:
        # TODO validate self.get_param('QueueUrl')
        attributes = self._get_param("Attributes", {})
        queue_name = self._get_queue_name()
        self.sqs_backend.set_queue_attributes(queue_name, attributes)
        return EmptyResult()

    def delete_queue(self) -> ActionResult:
        # TODO validate self.get_param('QueueUrl')
        queue_name = self._get_queue_name()
        self.sqs_backend.delete_queue(queue_name)
        return EmptyResult()

    def send_message(self) -> ActionResult:
        message = self._get_param("MessageBody")
        delay_seconds = self._get_param("DelaySeconds")
        message_group_id = self._get_param("MessageGroupId")
        message_dedupe_id = self._get_param("MessageDeduplicationId")
        message_attributes = self._get_param("MessageAttributes", {})
        validate_message_attributes(message_attributes)
        system_message_attributes = self._get_param("MessageSystemAttributes")
        validate_message_attributes(system_message_attributes)
        queue_name = self._get_queue_name()
        message = self.sqs_backend.send_message(
            queue_name,
            message,
            message_attributes=message_attributes,
            delay_seconds=delay_seconds,
            deduplication_id=message_dedupe_id,
            group_id=message_group_id,
            system_attributes=system_message_attributes,
        )
        resp = {
            "MD5OfMessageBody": message.body_md5,
            "MessageId": message.id,
        }
        if len(message.message_attributes) > 0:
            resp["MD5OfMessageAttributes"] = message.attribute_md5
        return ActionResult(resp)

    def send_message_batch(self) -> ActionResult:
        queue_name = self._get_queue_name()
        self.sqs_backend.get_queue(queue_name)
        entries = self._get_param("Entries", [])
        entries = {str(idx): entry for idx, entry in enumerate(entries)}
        for entry in entries.values():
            if "MessageAttributes" in entry:
                validate_message_attributes(entry["MessageAttributes"])
            else:
                entry["MessageAttributes"] = {}
            if "DelaySeconds" not in entry:
                entry["DelaySeconds"] = None

        if entries == {}:
            raise EmptyBatchRequest()

        messages, failed_invalid_delay = self.sqs_backend.send_message_batch(
            queue_name, entries
        )

        errors = []
        for entry in failed_invalid_delay:
            errors.append(
                {
                    "Id": entry["Id"],
                    "SenderFault": "true",
                    "Code": "InvalidParameterValue",
                    "Message": "Value 1800 for parameter DelaySeconds is invalid. Reason: DelaySeconds must be &gt;= 0 and &lt;= 900.",
                }
            )

        resp: dict[str, Any] = {"Successful": [], "Failed": errors}
        for msg in messages:
            msg_dict = {
                "Id": msg.user_id,  # type: ignore
                "MessageId": msg.id,
                "MD5OfMessageBody": msg.body_md5,
            }
            if len(msg.message_attributes) > 0:
                msg_dict["MD5OfMessageAttributes"] = msg.attribute_md5
            resp["Successful"].append(msg_dict)
        return ActionResult(resp)

    def delete_message(self) -> ActionResult:
        queue_name = self._get_queue_name()
        receipt_handle = self._get_param("ReceiptHandle")
        self.sqs_backend.delete_message(queue_name, receipt_handle)
        return EmptyResult()

    def delete_message_batch(self) -> ActionResult:
        queue_name = self._get_queue_name()
        receipts = self._get_param("Entries", [])
        if not receipts:
            raise EmptyBatchRequest(action="Delete")
        receipt_seen = set()
        for receipt_and_id in receipts:
            receipt = receipt_and_id["ReceiptHandle"]
            if receipt in receipt_seen:
                raise BatchEntryIdsNotDistinct(receipt_and_id["Id"])
            receipt_seen.add(receipt)
        success, failed = self.sqs_backend.delete_message_batch(queue_name, receipts)
        result = {"Successful": [{"Id": _id} for _id in success], "Failed": failed}
        return ActionResult(result)

    def purge_queue(self) -> ActionResult:
        queue_name = self._get_queue_name()
        self.sqs_backend.purge_queue(queue_name)
        return EmptyResult()

    def receive_message(self) -> ActionResult:
        queue_name = self._get_queue_name()
        attribute_names = self._get_param("AttributeNames", [])
        message_system_attribute_names = self._get_param(
            "MessageSystemAttributeNames", []
        )
        # AttributeNames has been deprecated in favor of MessageSystemAttributeNames.
        # We combine both here (with no duplicates) in order to support either parameter.
        message_system_attribute_names = list(
            set(attribute_names + message_system_attribute_names)
        )
        message_attribute_names = self._get_param("MessageAttributeNames", [])
        queue = self.sqs_backend.get_queue(queue_name)
        message_count = self._get_param(
            "MaxNumberOfMessages", DEFAULT_RECEIVED_MESSAGES
        )
        if message_count < 1 or message_count > 10:
            raise SQSException(
                "InvalidParameterValue",
                "An error occurred (InvalidParameterValue) when calling "
                f"the ReceiveMessage operation: Value {message_count} for parameter "
                "MaxNumberOfMessages is invalid. Reason: must be between "
                "1 and 10, if provided.",
            )
        try:
            wait_time = int(self._get_param("WaitTimeSeconds"))
        except TypeError:
            wait_time = int(queue.receive_message_wait_time_seconds)  # type: ignore
        if wait_time < 0 or wait_time > 20:
            raise SQSException(
                "InvalidParameterValue",
                "An error occurred (InvalidParameterValue) when calling "
                f"the ReceiveMessage operation: Value {wait_time} for parameter "
                "WaitTimeSeconds is invalid. Reason: must be &lt;= 0 and "
                "&gt;= 20 if provided.",
            )
        try:
            visibility_timeout = self._get_validated_visibility_timeout()
        except TypeError:
            visibility_timeout = queue.visibility_timeout  # type: ignore
        except ValueError:
            raise MaxVisibilityTimeout()

        messages = self.sqs_backend.receive_message(
            queue_name,
            message_count,
            wait_time,
            visibility_timeout,
            message_attribute_names,
        )
        SUPPORTED_SYSTEM_MESSAGE_ATTRIBUTE_NAMES = [
            "AWSTraceHeader",
            "ApproximateFirstReceiveTimestamp",
            "ApproximateReceiveCount",
            "MessageDeduplicationId",
            "MessageGroupId",
            "SenderId",
            "SentTimestamp",
            "SequenceNumber",
        ]
        attributes_to_include = []
        include_all = "All" in message_system_attribute_names
        for attribute_name in SUPPORTED_SYSTEM_MESSAGE_ATTRIBUTE_NAMES:
            if include_all or attribute_name in message_system_attribute_names:
                attributes_to_include.append(attribute_name)

        msgs = []
        for message in messages:
            msg: dict[str, Any] = {
                "MessageId": message.id,
                "ReceiptHandle": message.receipt_handle,
                "MD5OfBody": message.body_md5,
                "Body": message.body,
            }
            if len(message.message_attributes):
                msg["MessageAttributes"] = message.message_attributes
                msg["MD5OfMessageAttributes"] = message.attribute_md5
            attributes = {}
            if "SenderId" in attributes_to_include:
                attributes["SenderId"] = message.sender_id
            if "SentTimestamp" in attributes_to_include:
                attributes["SentTimestamp"] = str(message.sent_timestamp)
            if "ApproximateReceiveCount" in attributes_to_include:
                attributes["ApproximateReceiveCount"] = str(
                    message.approximate_receive_count
                )
            if "ApproximateFirstReceiveTimestamp" in attributes_to_include:
                attributes["ApproximateFirstReceiveTimestamp"] = str(
                    message.approximate_first_receive_timestamp
                )
            if "MessageDeduplicationId" in attributes_to_include:
                if message.deduplication_id is not None:
                    attributes["MessageDeduplicationId"] = message.deduplication_id
            if "MessageGroupId" in attributes_to_include:
                if message.group_id is not None:
                    attributes["MessageGroupId"] = message.group_id
            if "AWSTraceHeader" in attributes_to_include:
                if message.system_attributes and message.system_attributes.get(
                    "AWSTraceHeader"
                ):
                    attributes["AWSTraceHeader"] = message.system_attributes[
                        "AWSTraceHeader"
                    ].get("StringValue")
            if "SequenceNumber" in attributes_to_include:
                if message.sequence_number is not None:
                    attributes["SequenceNumber"] = message.sequence_number
            if attributes_to_include:
                msg["Attributes"] = attributes
            msgs.append(msg)

        result = {"Messages": msgs} if msgs else {}
        return ActionResult(result)

    def list_dead_letter_source_queues(self) -> ActionResult:
        request_url = urlparse(self.uri)
        queue_name = self._get_queue_name()
        queues = self.sqs_backend.list_dead_letter_source_queues(queue_name)
        result = {"queueUrls": [queue.url(request_url) for queue in queues]}
        return ActionResult(result)

    def add_permission(self) -> ActionResult:
        queue_name = self._get_queue_name()
        actions = self._get_param("Actions", [])
        account_ids = self._get_param("AWSAccountIds", [])
        label = self._get_param("Label")
        self.sqs_backend.add_permission(
            region_name=self.region,
            queue_name=queue_name,
            actions=actions,
            account_ids=account_ids,
            label=label,
        )
        return EmptyResult()

    def remove_permission(self) -> ActionResult:
        queue_name = self._get_queue_name()
        label = self._get_param("Label")
        self.sqs_backend.remove_permission(queue_name, label)
        return EmptyResult()

    def tag_queue(self) -> ActionResult:
        queue_name = self._get_queue_name()
        tags = self._get_param("Tags", [])
        self.sqs_backend.tag_queue(queue_name, tags)
        return EmptyResult()

    def untag_queue(self) -> ActionResult:
        queue_name = self._get_queue_name()
        tag_keys = self._get_param("TagKeys", [])
        self.sqs_backend.untag_queue(queue_name, tag_keys)
        return EmptyResult()

    def list_queue_tags(self) -> ActionResult:
        queue_name = self._get_queue_name()
        queue = self.sqs_backend.list_queue_tags(queue_name)
        result = {"Tags": queue.tags}
        return ActionResult(result)



"""Handles incoming pipes requests, invokes methods, returns responses."""

import json
from typing import Any
from urllib.parse import unquote

from moto.core.responses import BaseResponse
from moto.core.utils import iso_8601_datetime_without_milliseconds

from .exceptions import ValidationException
from .models import EventBridgePipesBackend, pipes_backends


class EventBridgePipesResponse(BaseResponse):
    """Handler for EventBridgePipes requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="pipes")

    @property
    def pipes_backend(self) -> EventBridgePipesBackend:
        return pipes_backends[self.current_account][self.region]

    def create_pipe(self) -> str:
        body_params = json.loads(self.body) if self.body else {}

        name = body_params.get("Name") or self.uri.split("/")[-1]
        description = body_params.get("Description")
        desired_state = body_params.get("DesiredState")
        source = body_params.get("Source")
        source_parameters = body_params.get("SourceParameters")
        enrichment = body_params.get("Enrichment")
        enrichment_parameters = body_params.get("EnrichmentParameters")
        target = body_params.get("Target")
        target_parameters = body_params.get("TargetParameters")
        role_arn = body_params.get("RoleArn")
        tags = body_params.get("Tags")
        log_configuration = body_params.get("LogConfiguration")
        kms_key_identifier = body_params.get("KmsKeyIdentifier")

        if not source:
            raise ValidationException("Source is a required parameter")
        if not target:
            raise ValidationException("Target is a required parameter")
        if not role_arn:
            raise ValidationException("RoleArn is a required parameter")

        pipe = self.pipes_backend.create_pipe(
            name=name,
            description=description,
            desired_state=desired_state,
            source=source,
            source_parameters=source_parameters,
            enrichment=enrichment,
            enrichment_parameters=enrichment_parameters,
            target=target,
            target_parameters=target_parameters,
            role_arn=role_arn,
            tags=tags,
            log_configuration=log_configuration,
            kms_key_identifier=kms_key_identifier,
        )
        return json.dumps(
            {
                "Arn": pipe.arn,
                "Name": pipe.name,
                "DesiredState": pipe.desired_state,
                "CurrentState": pipe.current_state,
                "CreationTime": iso_8601_datetime_without_milliseconds(
                    pipe.creation_time
                ),
                "LastModifiedTime": iso_8601_datetime_without_milliseconds(
                    pipe.last_modified_time
                ),
            }
        )

    def describe_pipe(self) -> str:
        name = self.uri.split("?")[0].split("/")[-1]
        pipe = self.pipes_backend.describe_pipe(
            name=name,
        )
        response_dict: dict[str, Any] = {
            "Arn": pipe.arn,
            "Name": pipe.name,
            "DesiredState": pipe.desired_state,
            "CurrentState": pipe.current_state,
            "CreationTime": iso_8601_datetime_without_milliseconds(pipe.creation_time),
            "LastModifiedTime": iso_8601_datetime_without_milliseconds(
                pipe.last_modified_time
            ),
        }
        if pipe.description is not None:
            response_dict["Description"] = pipe.description
        if pipe.state_reason is not None:
            response_dict["StateReason"] = pipe.state_reason
        if pipe.source is not None:
            response_dict["Source"] = pipe.source
        if pipe.source_parameters is not None:
            response_dict["SourceParameters"] = pipe.source_parameters
        if pipe.enrichment is not None:
            response_dict["Enrichment"] = pipe.enrichment
        if pipe.enrichment_parameters is not None:
            response_dict["EnrichmentParameters"] = pipe.enrichment_parameters
        if pipe.target is not None:
            response_dict["Target"] = pipe.target
        if pipe.target_parameters is not None:
            response_dict["TargetParameters"] = pipe.target_parameters
        if pipe.role_arn is not None:
            response_dict["RoleArn"] = pipe.role_arn
        if pipe.tags is not None:
            response_dict["Tags"] = pipe.tags
        if pipe.log_configuration is not None:
            response_dict["LogConfiguration"] = pipe.log_configuration
        if pipe.kms_key_identifier is not None:
            response_dict["KmsKeyIdentifier"] = pipe.kms_key_identifier
        return json.dumps(response_dict)

    def delete_pipe(self) -> str:
        name = self.uri.split("?")[0].split("/")[-1]

        arn, name, desired_state, current_state, creation_time, last_modified_time = (
            self.pipes_backend.delete_pipe(
                name=name,
            )
        )

        return json.dumps(
            {
                "Arn": arn,
                "Name": name,
                "DesiredState": desired_state,
                "CurrentState": current_state,
                "CreationTime": creation_time,
                "LastModifiedTime": last_modified_time,
            }
        )

    def tag_resource(self) -> str:
        resource_arn = unquote(self.uri.split("/tags/")[-1])
        body_params = json.loads(self.body) if self.body else {}
        tags = body_params.get("Tags") or body_params.get("tags")
        if not tags:
            raise ValidationException("Tags is a required parameter")

        self.pipes_backend.tag_resource(
            resource_arn=resource_arn,
            tags=tags,
        )
        return json.dumps({})

    def untag_resource(self) -> str:
        resource_arn = unquote(self.uri.split("?")[0].split("/tags/")[-1])
        tag_keys = self.querystring.get("tagKeys", [])

        self.pipes_backend.untag_resource(
            resource_arn=resource_arn,
            tag_keys=tag_keys,
        )
        return json.dumps({})

    def list_pipes(self) -> str:
        params = json.loads(self.body) if self.body else {}
        if not params and self.querystring:
            params = {
                k: (v[0] if isinstance(v, list) and len(v) > 0 else v)
                for k, v in self.querystring.items()
            }

        name_prefix = params.get("NamePrefix") or params.get("namePrefix")
        desired_state = params.get("DesiredState") or params.get("desiredState")
        current_state = params.get("CurrentState") or params.get("currentState")
        source_prefix = params.get("SourcePrefix") or params.get("sourcePrefix")
        target_prefix = params.get("TargetPrefix") or params.get("targetPrefix")
        next_token = params.get("NextToken") or params.get("nextToken")
        limit = params.get("Limit") or params.get("limit")
        if limit is not None and limit != "":
            try:
                limit = int(limit)
            except (ValueError, TypeError):
                limit = None

        pipes, next_token = self.pipes_backend.list_pipes(
            name_prefix=name_prefix,
            desired_state=desired_state,
            current_state=current_state,
            source_prefix=source_prefix,
            target_prefix=target_prefix,
            next_token=next_token,
            limit=limit,
        )

        response_dict: dict[str, Any] = {"Pipes": pipes}
        if next_token:
            response_dict["NextToken"] = next_token

        return json.dumps(response_dict)



from __future__ import annotations

import copy
import logging
import threading
from typing import Any, Final, Optional

from moto.stepfunctions.parser.api import (
    Arn,
    ExecutionFailedEventDetails,
    StateMachineType,
    Timestamp,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.itemprocessor.map_run_record import (
    MapRunRecordPoolManager,
)
from moto.stepfunctions.parser.asl.eval.callback.callback import CallbackPoolManager
from moto.stepfunctions.parser.asl.eval.evaluation_details import AWSExecutionDetails
from moto.stepfunctions.parser.asl.eval.event.event_manager import (
    EventHistoryContext,
    EventManager,
)
from moto.stepfunctions.parser.asl.eval.event.logging import (
    CloudWatchLoggingSession,
)
from moto.stepfunctions.parser.asl.eval.program_state import (
    ProgramEnded,
    ProgramError,
    ProgramRunning,
    ProgramState,
    ProgramStopped,
    ProgramTimedOut,
)
from moto.stepfunctions.parser.asl.eval.states import ContextObjectData, States
from moto.stepfunctions.parser.asl.eval.variable_store import VariableStore
from moto.stepfunctions.parser.backend.activity import Activity
from moto.stepfunctions.parser.mocking.mock_config import MockedResponse, MockTestCase

LOG = logging.getLogger(__name__)


class Environment:
    _state_mutex: Final[threading.RLock()]
    _program_state: Optional[ProgramState]
    program_state_event: Final[threading.Event()]

    event_manager: EventManager
    event_history_context: Final[EventHistoryContext]
    cloud_watch_logging_session: Final[Optional[CloudWatchLoggingSession]]
    aws_execution_details: Final[AWSExecutionDetails]
    execution_type: Final[StateMachineType]
    callback_pool_manager: CallbackPoolManager
    map_run_record_pool_manager: MapRunRecordPoolManager
    activity_store: Final[dict[Arn, Activity]]
    mock_test_case: Optional[MockTestCase] = None

    _frames: Final[list[Environment]]
    _is_frame: bool = False

    heap: dict[str, Any] = {}
    stack: list[Any] = []
    states: Final[States]
    variable_store: Final[VariableStore]

    def __init__(
        self,
        aws_execution_details: AWSExecutionDetails,
        execution_type: StateMachineType,
        context: ContextObjectData,
        event_history_context: EventHistoryContext,
        cloud_watch_logging_session: Optional[CloudWatchLoggingSession],
        activity_store: dict[Arn, Activity],
        variable_store: Optional[VariableStore] = None,
        mock_test_case: Optional[MockTestCase] = None,
    ):
        super().__init__()
        self._state_mutex = threading.RLock()
        self._program_state = None
        self.program_state_event = threading.Event()

        self.cloud_watch_logging_session = cloud_watch_logging_session
        self.event_manager = EventManager(
            cloud_watch_logging_session=cloud_watch_logging_session
        )
        self.event_history_context = event_history_context

        self.aws_execution_details = aws_execution_details
        self.execution_type = execution_type
        self.callback_pool_manager = CallbackPoolManager(activity_store=activity_store)
        self.map_run_record_pool_manager = MapRunRecordPoolManager()

        self.activity_store = activity_store

        self.mock_test_case = mock_test_case

        self._frames = []
        self._is_frame = False

        self.heap = {}
        self.stack = []
        self.states = States(context=context)
        self.variable_store = variable_store or VariableStore()

    @classmethod
    def as_frame_of(
        cls,
        env: Environment,
        event_history_frame_cache: Optional[EventHistoryContext] = None,
    ) -> Environment:
        return Environment.as_inner_frame_of(
            env=env,
            variable_store=env.variable_store,
            event_history_frame_cache=event_history_frame_cache,
        )

    @classmethod
    def as_inner_frame_of(
        cls,
        env: Environment,
        variable_store: VariableStore,
        event_history_frame_cache: Optional[EventHistoryContext] = None,
    ) -> Environment:
        # Construct the frame's context object data.
        context = ContextObjectData(
            Execution=env.states.context_object.context_object_data["Execution"],
            StateMachine=env.states.context_object.context_object_data["StateMachine"],
        )
        if "Task" in env.states.context_object.context_object_data:
            context["Task"] = env.states.context_object.context_object_data["Task"]

        # The default logic provisions for child frame to extend the source frame event id.
        if event_history_frame_cache is None:
            event_history_frame_cache = EventHistoryContext(
                previous_event_id=env.event_history_context.source_event_id
            )

        frame = cls(
            aws_execution_details=env.aws_execution_details,
            execution_type=env.execution_type,
            context=context,
            event_history_context=event_history_frame_cache,
            cloud_watch_logging_session=env.cloud_watch_logging_session,
            activity_store=env.activity_store,
            variable_store=variable_store,
            mock_test_case=env.mock_test_case,
        )
        frame._is_frame = True
        frame.event_manager = env.event_manager
        if "State" in env.states.context_object.context_object_data:
            frame.states.context_object.context_object_data["State"] = copy.deepcopy(
                env.states.context_object.context_object_data["State"]
            )
        frame.callback_pool_manager = env.callback_pool_manager
        frame.map_run_record_pool_manager = env.map_run_record_pool_manager
        frame.heap = {}
        frame._program_state = copy.deepcopy(env._program_state)
        return frame

    @property
    def next_state_name(self) -> Optional[str]:
        next_state_name: Optional[str] = None
        program_state = self._program_state
        if isinstance(program_state, ProgramRunning):
            next_state_name = program_state.next_state_name
        return next_state_name

    @next_state_name.setter
    def next_state_name(self, next_state_name: str) -> None:
        if self._program_state is None:
            self._program_state = ProgramRunning()

        if isinstance(self._program_state, ProgramRunning):
            self._program_state.next_state_name = next_state_name
        else:
            raise RuntimeError(
                f"Could not set NextState value when in state '{type(self._program_state)}'."
            )

    @property
    def next_field_name(self) -> Optional[str]:
        next_field_name: Optional[str] = None
        program_state = self._program_state
        if isinstance(program_state, ProgramRunning):
            next_field_name = program_state.next_field_name
        return next_field_name

    @next_field_name.setter
    def next_field_name(self, next_field_name: str) -> None:
        if isinstance(self._program_state, ProgramRunning):
            self._program_state.next_field_name = next_field_name
        else:
            raise RuntimeError(
                f"Could not set NextField value when in state '{type(self._program_state)}'."
            )

    def program_state(self) -> ProgramState:
        return copy.deepcopy(self._program_state)

    def is_running(self) -> bool:
        return isinstance(self._program_state, ProgramRunning)

    def set_ended(self) -> None:
        with self._state_mutex:
            if isinstance(self._program_state, ProgramRunning):
                self._program_state = ProgramEnded()
                for frame in self._frames:
                    frame.set_ended()
            self.program_state_event.set()
            self.program_state_event.clear()

    def set_error(self, error: ExecutionFailedEventDetails) -> None:
        with self._state_mutex:
            self._program_state = ProgramError(error=error)
            for frame in self._frames:
                frame.set_error(error=error)
            self.program_state_event.set()
            self.program_state_event.clear()

    def set_timed_out(self) -> None:
        with self._state_mutex:
            self._program_state = ProgramTimedOut()
            for frame in self._frames:
                frame.set_timed_out()
            self.program_state_event.set()
            self.program_state_event.clear()

    def set_stop(
        self, stop_date: Timestamp, cause: Optional[str], error: Optional[str]
    ) -> None:
        with self._state_mutex:
            if isinstance(self._program_state, ProgramRunning):
                self._program_state = ProgramStopped(
                    stop_date=stop_date, cause=cause, error=error
                )
                for frame in self._frames:
                    frame.set_stop(stop_date=stop_date, cause=cause, error=error)
                self.program_state_event.set()
                self.program_state_event.clear()

    def open_frame(
        self, event_history_context: Optional[EventHistoryContext] = None
    ) -> Environment:
        with self._state_mutex:
            frame = self.as_frame_of(
                env=self, event_history_frame_cache=event_history_context
            )
            self._frames.append(frame)
            return frame

    def open_inner_frame(
        self, event_history_context: Optional[EventHistoryContext] = None
    ) -> Environment:
        with self._state_mutex:
            variable_store = VariableStore.as_inner_scope_of(
                outer_variable_store=self.variable_store
            )
            frame = self.as_inner_frame_of(
                env=self,
                variable_store=variable_store,
                event_history_frame_cache=event_history_context,
            )
            self._frames.append(frame)
            return frame

    def close_frame(self, frame: Environment) -> None:
        with self._state_mutex:
            if frame in self._frames:
                self._frames.remove(frame)
                self.event_history_context.integrate(frame.event_history_context)

    def delete_frame(self, frame: Environment) -> None:
        with self._state_mutex:
            if frame in self._frames:
                self._frames.remove(frame)

    def is_frame(self) -> bool:
        return self._is_frame

    def is_standard_workflow(self) -> bool:
        return self.execution_type == StateMachineType.STANDARD

    def is_mocked_mode(self) -> bool:
        """
        Returns True if the state machine is running in mock mode and the current
        state has a defined mock configuration in the target environment or frame;
        otherwise, returns False.
        """
        return (
            self.mock_test_case is not None
            and self.next_state_name in self.mock_test_case.state_mocked_responses
        )

    def get_current_mocked_response(self) -> MockedResponse:
        if not self.is_mocked_mode():
            raise RuntimeError(
                "Cannot retrieve mocked response: execution is not operating in mocked mode"
            )
        state_name = self.next_state_name
        state_mocked_responses: Optional = (
            self.mock_test_case.state_mocked_responses.get(state_name)
        )
        if state_mocked_responses is None:
            raise RuntimeError(
                f"No mocked response definition for state '{state_name}'"
            )
        retry_count = self.states.context_object.context_object_data["State"][
            "RetryCount"
        ]
        if len(state_mocked_responses.mocked_responses) <= retry_count:
            raise RuntimeError(
                f"No mocked response definition for state '{state_name}' "
                f"and retry number '{retry_count}'"
            )
        return state_mocked_responses.mocked_responses[retry_count]



from __future__ import annotations

import logging
from typing import Final

from moto.stepfunctions.parser import analytics
from moto.stepfunctions.parser.asl.antlr.runtime.ASLParser import ASLParser
from moto.stepfunctions.parser.asl.component.common.query_language import (
    QueryLanguageMode,
)
from moto.stepfunctions.parser.asl.static_analyser.static_analyser import StaticAnalyser

LOG = logging.getLogger(__name__)


class QueryLanguage(str):
    JSONPath = QueryLanguageMode.JSONPath.name
    JSONata = QueryLanguageMode.JSONata.name
    Both = "JSONPath+JSONata"


class UsageMetricsStaticAnalyser(StaticAnalyser):
    @staticmethod
    def process(definition: str) -> UsageMetricsStaticAnalyser:
        analyser = UsageMetricsStaticAnalyser()
        try:
            # Run the static analyser.
            analyser.analyse(definition=definition)

            # Determine which query language is being used in this state machine.
            query_modes = analyser.query_language_modes
            if len(query_modes) == 2:
                language_used = QueryLanguage.Both
            elif QueryLanguageMode.JSONata in query_modes:
                language_used = QueryLanguage.JSONata
            else:
                language_used = QueryLanguage.JSONPath

            # Determine is the state machine uses the variables feature.
            uses_variables = analyser.uses_variables

            # Count.
            analytics.language_features_counter.labels(
                query_language=language_used, uses_variables=uses_variables
            ).increment()
        except Exception as e:
            LOG.warning(
                "Failed to record Step Functions metrics from static analysis",
                exc_info=e,
            )
        return analyser

    query_language_modes: Final[set[QueryLanguageMode]]
    uses_variables: bool

    def __init__(self):
        super().__init__()
        self.query_language_modes = set()
        self.uses_variables = False

    def visitQuery_language_decl(self, ctx: ASLParser.Query_language_declContext):
        if len(self.query_language_modes) == 2:
            # Both query language modes have been confirmed to be in use.
            return
        query_language_mode_int = ctx.children[-1].getSymbol().type
        query_language_mode = QueryLanguageMode(value=query_language_mode_int)
        self.query_language_modes.add(query_language_mode)

    def visitState_decl(self, ctx: ASLParser.State_declContext):
        # If before entering a state, no query language was explicitly enforced, then we know
        # this is the first state operating under the default mode (JSONPath)
        if not self.query_language_modes:
            self.query_language_modes.add(QueryLanguageMode.JSONPath)
        super().visitState_decl(ctx=ctx)

    def visitString_literal(self, ctx: ASLParser.String_literalContext):
        # Prune everything parsed as a string literal.
        return

    def visitString_variable_sample(self, ctx: ASLParser.String_variable_sampleContext):
        self.uses_variables = True

    def visitAssign_decl(self, ctx: ASLParser.Assign_declContext):
        self.uses_variables = True



from typing import Final

from moto.stepfunctions.parser.asl.component.common.assign.assign_template_value_object import (
    AssignTemplateValueObject,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment


class ItemSelector(EvalComponent):
    template_value_object: Final[AssignTemplateValueObject]

    def __init__(self, template_value_object: AssignTemplateValueObject):
        self.template_value_object = template_value_object

    def _eval_body(self, env: Environment) -> None:
        self.template_value_object.eval(env=env)
